# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: sdv/databroker/v1/collector.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'sdv/databroker/v1/collector.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sdv.databroker.v1 import types_pb2 as sdv_dot_databroker_dot_v1_dot_types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!sdv/databroker/v1/collector.proto\x12\x11sdv.databroker.v1\x1a\x1dsdv/databroker/v1/types.proto\"\xba\x01\n\x17UpdateDatapointsRequest\x12N\n\ndatapoints\x18\x01 \x03(\x0b\x32:.sdv.databroker.v1.UpdateDatapointsRequest.DatapointsEntry\x1aO\n\x0f\x44\x61tapointsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.sdv.databroker.v1.Datapoint:\x02\x38\x01\"\xaf\x01\n\x15UpdateDatapointsReply\x12\x44\n\x06\x65rrors\x18\x01 \x03(\x0b\x32\x34.sdv.databroker.v1.UpdateDatapointsReply.ErrorsEntry\x1aP\n\x0b\x45rrorsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\x30\n\x05value\x18\x02 \x01(\x0e\x32!.sdv.databroker.v1.DatapointError:\x02\x38\x01\"\xba\x01\n\x17StreamDatapointsRequest\x12N\n\ndatapoints\x18\x01 \x03(\x0b\x32:.sdv.databroker.v1.StreamDatapointsRequest.DatapointsEntry\x1aO\n\x0f\x44\x61tapointsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.sdv.databroker.v1.Datapoint:\x02\x38\x01\"\xaf\x01\n\x15StreamDatapointsReply\x12\x44\n\x06\x65rrors\x18\x01 \x03(\x0b\x32\x34.sdv.databroker.v1.StreamDatapointsReply.ErrorsEntry\x1aP\n\x0b\x45rrorsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\x30\n\x05value\x18\x02 \x01(\x0e\x32!.sdv.databroker.v1.DatapointError:\x02\x38\x01\"R\n\x19RegisterDatapointsRequest\x12\x35\n\x04list\x18\x01 \x03(\x0b\x32\'.sdv.databroker.v1.RegistrationMetadata\"\x9d\x01\n\x14RegistrationMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12.\n\tdata_type\x18\x02 \x01(\x0e\x32\x1b.sdv.databroker.v1.DataType\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x32\n\x0b\x63hange_type\x18\x04 \x01(\x0e\x32\x1d.sdv.databroker.v1.ChangeType\"\x93\x01\n\x17RegisterDatapointsReply\x12H\n\x07results\x18\x01 \x03(\x0b\x32\x37.sdv.databroker.v1.RegisterDatapointsReply.ResultsEntry\x1a.\n\x0cResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x32\xd3\x02\n\tCollector\x12n\n\x12RegisterDatapoints\x12,.sdv.databroker.v1.RegisterDatapointsRequest\x1a*.sdv.databroker.v1.RegisterDatapointsReply\x12h\n\x10UpdateDatapoints\x12*.sdv.databroker.v1.UpdateDatapointsRequest\x1a(.sdv.databroker.v1.UpdateDatapointsReply\x12l\n\x10StreamDatapoints\x12*.sdv.databroker.v1.StreamDatapointsRequest\x1a(.sdv.databroker.v1.StreamDatapointsReply(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sdv.databroker.v1.collector_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_UPDATEDATAPOINTSREQUEST_DATAPOINTSENTRY']._loaded_options = None
  _globals['_UPDATEDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_options = b'8\001'
  _globals['_UPDATEDATAPOINTSREPLY_ERRORSENTRY']._loaded_options = None
  _globals['_UPDATEDATAPOINTSREPLY_ERRORSENTRY']._serialized_options = b'8\001'
  _globals['_STREAMDATAPOINTSREQUEST_DATAPOINTSENTRY']._loaded_options = None
  _globals['_STREAMDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_options = b'8\001'
  _globals['_STREAMDATAPOINTSREPLY_ERRORSENTRY']._loaded_options = None
  _globals['_STREAMDATAPOINTSREPLY_ERRORSENTRY']._serialized_options = b'8\001'
  _globals['_REGISTERDATAPOINTSREPLY_RESULTSENTRY']._loaded_options = None
  _globals['_REGISTERDATAPOINTSREPLY_RESULTSENTRY']._serialized_options = b'8\001'
  _globals['_UPDATEDATAPOINTSREQUEST']._serialized_start=88
  _globals['_UPDATEDATAPOINTSREQUEST']._serialized_end=274
  _globals['_UPDATEDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_start=195
  _globals['_UPDATEDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_end=274
  _globals['_UPDATEDATAPOINTSREPLY']._serialized_start=277
  _globals['_UPDATEDATAPOINTSREPLY']._serialized_end=452
  _globals['_UPDATEDATAPOINTSREPLY_ERRORSENTRY']._serialized_start=372
  _globals['_UPDATEDATAPOINTSREPLY_ERRORSENTRY']._serialized_end=452
  _globals['_STREAMDATAPOINTSREQUEST']._serialized_start=455
  _globals['_STREAMDATAPOINTSREQUEST']._serialized_end=641
  _globals['_STREAMDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_start=195
  _globals['_STREAMDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_end=274
  _globals['_STREAMDATAPOINTSREPLY']._serialized_start=644
  _globals['_STREAMDATAPOINTSREPLY']._serialized_end=819
  _globals['_STREAMDATAPOINTSREPLY_ERRORSENTRY']._serialized_start=372
  _globals['_STREAMDATAPOINTSREPLY_ERRORSENTRY']._serialized_end=452
  _globals['_REGISTERDATAPOINTSREQUEST']._serialized_start=821
  _globals['_REGISTERDATAPOINTSREQUEST']._serialized_end=903
  _globals['_REGISTRATIONMETADATA']._serialized_start=906
  _globals['_REGISTRATIONMETADATA']._serialized_end=1063
  _globals['_REGISTERDATAPOINTSREPLY']._serialized_start=1066
  _globals['_REGISTERDATAPOINTSREPLY']._serialized_end=1213
  _globals['_REGISTERDATAPOINTSREPLY_RESULTSENTRY']._serialized_start=1167
  _globals['_REGISTERDATAPOINTSREPLY_RESULTSENTRY']._serialized_end=1213
  _globals['_COLLECTOR']._serialized_start=1216
  _globals['_COLLECTOR']._serialized_end=1555
# @@protoc_insertion_point(module_scope)



#! /usr/bin/env python

########################################################################
# Copyright (c) 2020 Robert Bosch GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
########################################################################

import asyncio
import threading
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional

from . import cli_backend


class KuksaClientThread(threading.Thread):

    # Constructor
    def __init__(self, config):
        super().__init__()

        self.backend = cli_backend.Backend.from_config(config)
        self.loop = None

    # PEP 702 deprecated decorator available first in Python 3.13
    def checkConnection(self) -> bool:
        """
        Check if thread has established a connection to the broker/server.
        Note that this method does not indicate the current state of the connection,
        This method may return True even if the broker/server currently is not reachable.
        Deprecated - Use connection_established() instead"
        """
        return self.connection_established()

    def connection_established(self) -> bool:
        """
        Check if thread has established a connection to the broker/server.
        Note that this method does not indicate the current state of the connection,
        This method may return True even if the broker/server currently is not reachable.
        """
        return self.backend.connection_established()

    def stop(self):
        self.backend.stop()

    # Do authorization by passing a jwt token or a token file
    def authorize(self, token_or_tokenfile: Optional[str] = None, timeout=5):
        return self.backend.authorize(token_or_tokenfile, timeout)

    # Update VSS Tree Entry
    def updateVSSTree(self, jsonStr: str, timeout=5):
        return self.backend.updateVSSTree(jsonStr, timeout)

    # Update Meta Data of a given path
    def updateMetaData(self, path: str, jsonStr: str, timeout=5):
        return self.backend.updateMetaData(path, jsonStr, timeout)

    # Get Meta Data of a given path
    def getMetaData(self, path: str, timeout=5):
        return self.backend.getMetaData(path, timeout)

    # Set value to a given path
    def setValue(self, path: str, value, attribute="value", timeout=5):
        return self.backend.setValue(path, value, attribute, timeout)

    # Set value of every given path
    def setValues(self, updates: Dict[str, Any], attribute="value", timeout=5):
        return self.backend.setValues(updates, attribute, timeout)

    # Get value to a given path
    def getValue(self, path: str, attribute="value", timeout=5):
        return self.backend.getValue(path, attribute, timeout)

    # Get value of every given path
    def getValues(self, paths: Iterable[str], attribute="value", timeout=5):
        return self.backend.getValues(paths, attribute, timeout)

    # Subscribe value changes of to a given path.
    # The given callback function will be called then, if the given path is updated:
    #   updateMessage = await webSocket.recv()
    #   callback(updateMessage)
    def subscribe(self, path: str, callback, attribute="value", timeout=5):
        return self.backend.subscribe(path, callback, attribute, timeout)

    def subscribeMultiple(self, paths: Iterable[str], callback, attribute="value", timeout=5):
        return self.backend.subscribeMultiple(paths, callback, attribute, timeout)

    # Unsubscribe value changes of to a given path.
    # The subscription id from the response of the corresponding subscription request will be required
    def unsubscribe(self, sub_id: int, timeout=5):
        return self.backend.unsubscribe(sub_id, timeout)

    def disconnect(self, timeout=5):
        return self.backend.disconnect(timeout)

    def connect(self, timeout=5):
        return self.backend.connect(timeout)

    # Thread function: Start the asyncio loop
    def run(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self.backend.mainLoop())



# /********************************************************************************
# * Copyright (c) 2023 Contributors to the Eclipse Foundation
# *
# * See the NOTICE file(s) distributed with this work for additional
# * information regarding copyright ownership.
# *
# * This program and the accompanying materials are made available under the
# * terms of the Apache License 2.0 which is available at
# * http://www.apache.org/licenses/LICENSE-2.0
# *
# * SPDX-License-Identifier: Apache-2.0
# ********************************************************************************/

import pytest
from kuksa_client.grpc import Datapoint
from kuksa.val.v1 import types_pb2
from google.protobuf import timestamp_pb2

#
# Client rules:
# For simple strings like abd it is optional to quote them ("abc") or not (abc)
# Quotes are needed if you have commas ("ab, c")
# If you have duoble quotes in strings you must escape them
#
# Note that KUKSA Server has different rules, there the payload must be valid JSON,
# so not all tests shown below are recommended as they cannot be used for KUKSA Server


def test_array_parse_no_quote():
    """
    No need for quotes just because you have a blank
    """
    test_str = r'[say hello, abc]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == "say hello"
    assert my_array[1] == "abc"


def test_array_parse_no_inside_quote():
    """Quotes are OK"""
    test_str = r'["say hello","abc"]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == "say hello"
    assert my_array[1] == "abc"


def test_array_parse_no_inside_quote_single():
    """Quotes are OK"""
    test_str = "['say hello','abc']"
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == "say hello"
    assert my_array[1] == "abc"


def test_array_parse_double_quote():
    test_str = r'["say \"hello\"","abc"]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == "say \"hello\""
    assert my_array[1] == "abc"


def test_array_parse_single_quote():
    test_str = r'[say \'hello\',abc]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == "say 'hello'"
    assert my_array[1] == "abc"


def test_array_parse_comma():
    test_str = r'["say, hello","abc"]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == r'say, hello'
    assert my_array[1] == "abc"


def test_array_square():
    """No problem having square brackets as part of strings"""
    test_str = r'[say hello[], abc]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == "say hello[]"
    assert my_array[1] == "abc"


def test_array_empty_string_quoted():
    test_str = r'["", abc]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == ""
    assert my_array[1] == "abc"


def test_array_empty_string_not_quoted():
    """In this case the first item is ignored"""
    test_str = r'[, abc]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 1
    assert my_array[0] == "abc"


def test_double_comma():
    """In this case the middle item is ignored"""
    test_str = r'[def,, abc]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 2
    assert my_array[0] == "def"
    assert my_array[1] == "abc"


def test_quotes_in_string_values():
    """Escaped double quotes, so in total 4 items"""
    test_str = r'["dtc1, dtc2", dtc3, \" dtc4, dtc4\"]'
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 4
    assert my_array[0] == "dtc1, dtc2"
    assert my_array[1] == "dtc3"
    assert my_array[2] == "\" dtc4"
    assert my_array[3] == "dtc4\""


def test_quotes_in_string_values_2():
    """Double quotes in double quotes so in total three values"""
    test_str = "['dtc1, dtc2', dtc3, \" dtc4, dtc4\"]"
    my_array = list(Datapoint.cast_array_values(Datapoint.cast_str, test_str))
    assert len(my_array) == 3
    assert my_array[0] == 'dtc1, dtc2'
    assert my_array[1] == "dtc3"
    assert my_array[2] == " dtc4, dtc4"


def test_int_no_quote():
    test_str = r'[123,456]'
    my_array = list(Datapoint.cast_array_values(int, test_str))
    assert len(my_array) == 2
    assert my_array[0] == 123
    assert my_array[1] == 456


def test_int_quote():
    """Quoting individual int values is not allowed"""
    test_str = r'["123","456"]'
    with pytest.raises(ValueError):
        list(Datapoint.cast_array_values(int, test_str))


def test_float_no_quote():
    test_str = r'[123,456.23]'
    my_array = list(Datapoint.cast_array_values(float, test_str))
    assert len(my_array) == 2
    assert my_array[0] == 123
    assert my_array[1] == 456.23


def test_cast_str():
    """Unquoted quotation marks shall be removed, quoted kept without quotes"""
    test_str = r'"say hello"'
    assert Datapoint.cast_str(test_str) == r'say hello'
    test_str = r'"say \"hello\""'
    assert Datapoint.cast_str(test_str) == r'say "hello"'
    test_str = r'say "hello"'
    assert Datapoint.cast_str(test_str) == r'say "hello"'


def test_cast_bool():
    assert Datapoint.cast_bool("true") is True
    assert Datapoint.cast_bool("True") is True
    assert Datapoint.cast_bool("T") is True
    assert Datapoint.cast_bool("t") is True
    assert Datapoint.cast_bool("false") is False
    assert Datapoint.cast_bool("False") is False
    assert Datapoint.cast_bool("F") is False
    assert Datapoint.cast_bool("f") is False

    # And then some other, treated as true for now
    assert Datapoint.cast_bool("Ja") is True
    assert Datapoint.cast_bool("Nein") is True
    assert Datapoint.cast_bool("Doch") is True


def test_from_message_none():
    """
    There shall always be a value
    """
    msg = types_pb2.Datapoint()
    datapoint = Datapoint.from_message(msg)
    assert datapoint is None


def test_from_message_uint32():
    msg = types_pb2.Datapoint(uint32=456)
    datapoint = Datapoint.from_message(msg)
    assert datapoint.value == 456


def test_from_message_time():
    """
    Make sure that we can handle values out of range (by discarding them)
    gRPC supports year up to including 9999, so any date in year 10000 or later shall
    result in that None is returned
    """
    # Wed Jan 17 2024 10:02:27 GMT+0000
    timestamp = timestamp_pb2.Timestamp(seconds=1705485747)
    msg = types_pb2.Datapoint(uint32=456, timestamp=timestamp)
    datapoint = Datapoint.from_message(msg)
    assert datapoint.timestamp.year == 2024

    # Thu Dec 30 9999 07:13:22 GMT+0000
    timestamp = timestamp_pb2.Timestamp(seconds=253402154002)
    msg = types_pb2.Datapoint(uint32=456, timestamp=timestamp)
    datapoint = Datapoint.from_message(msg)
    assert datapoint.timestamp.year == 9999

    # Sat Jan 29 10000 07:13:22 GMT+0000
    # Currently the constructors does not check range
    timestamp = timestamp_pb2.Timestamp(seconds=253404746002)
    msg = types_pb2.Datapoint(uint32=456, timestamp=timestamp)

    # But the from_message method handle the checks
    datapoint = Datapoint.from_message(msg)
    assert datapoint is None



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from kuksa.val.v1 import val_pb2 as kuksa_dot_val_dot_v1_dot_val__pb2

GRPC_GENERATED_VERSION = '1.68.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in kuksa/val/v1/val_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class VALStub(object):
    """Note on authorization:
    Tokens (auth-token or auth-uuid) are sent as (GRPC / http2) metadata.

    The auth-token is a JWT compliant token as the examples found here:
    https://github.com/eclipse-kuksa/kuksa-databroker/tree/main/certificates/jwt

    See also https://github.com/eclipse-kuksa/kuksa-databroker/blob/main/doc/authorization.md#jwt-access-token

    Upon reception of auth-token, server shall generate an auth-uuid in metadata
    that the client can use instead of auth-token in subsequent calls.

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Get = channel.unary_unary(
                '/kuksa.val.v1.VAL/Get',
                request_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.GetRequest.SerializeToString,
                response_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.GetResponse.FromString,
                _registered_method=True)
        self.Set = channel.unary_unary(
                '/kuksa.val.v1.VAL/Set',
                request_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.SetRequest.SerializeToString,
                response_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.SetResponse.FromString,
                _registered_method=True)
        self.StreamedUpdate = channel.stream_stream(
                '/kuksa.val.v1.VAL/StreamedUpdate',
                request_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.StreamedUpdateRequest.SerializeToString,
                response_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.StreamedUpdateResponse.FromString,
                _registered_method=True)
        self.Subscribe = channel.unary_stream(
                '/kuksa.val.v1.VAL/Subscribe',
                request_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.SubscribeRequest.SerializeToString,
                response_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.SubscribeResponse.FromString,
                _registered_method=True)
        self.GetServerInfo = channel.unary_unary(
                '/kuksa.val.v1.VAL/GetServerInfo',
                request_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.GetServerInfoRequest.SerializeToString,
                response_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.GetServerInfoResponse.FromString,
                _registered_method=True)


class VALServicer(object):
    """Note on authorization:
    Tokens (auth-token or auth-uuid) are sent as (GRPC / http2) metadata.

    The auth-token is a JWT compliant token as the examples found here:
    https://github.com/eclipse-kuksa/kuksa-databroker/tree/main/certificates/jwt

    See also https://github.com/eclipse-kuksa/kuksa-databroker/blob/main/doc/authorization.md#jwt-access-token

    Upon reception of auth-token, server shall generate an auth-uuid in metadata
    that the client can use instead of auth-token in subsequent calls.

    """

    def Get(self, request, context):
        """Get entries
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Set(self, request, context):
        """Set entries
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamedUpdate(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Subscribe(self, request, context):
        """Subscribe to a set of entries

        Returns a stream of notifications.

        InvalidArgument is returned if the request is malformed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetServerInfo(self, request, context):
        """Shall return information that allows the client to determine
        what server/server implementation/version it is talking to
        eg. kuksa-databroker 0.5.1
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_VALServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Get': grpc.unary_unary_rpc_method_handler(
                    servicer.Get,
                    request_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.GetRequest.FromString,
                    response_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.GetResponse.SerializeToString,
            ),
            'Set': grpc.unary_unary_rpc_method_handler(
                    servicer.Set,
                    request_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.SetRequest.FromString,
                    response_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.SetResponse.SerializeToString,
            ),
            'StreamedUpdate': grpc.stream_stream_rpc_method_handler(
                    servicer.StreamedUpdate,
                    request_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.StreamedUpdateRequest.FromString,
                    response_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.StreamedUpdateResponse.SerializeToString,
            ),
            'Subscribe': grpc.unary_stream_rpc_method_handler(
                    servicer.Subscribe,
                    request_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.SubscribeRequest.FromString,
                    response_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.SubscribeResponse.SerializeToString,
            ),
            'GetServerInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetServerInfo,
                    request_deserializer=kuksa_dot_val_dot_v1_dot_val__pb2.GetServerInfoRequest.FromString,
                    response_serializer=kuksa_dot_val_dot_v1_dot_val__pb2.GetServerInfoResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'kuksa.val.v1.VAL', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('kuksa.val.v1.VAL', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class VAL(object):
    """Note on authorization:
    Tokens (auth-token or auth-uuid) are sent as (GRPC / http2) metadata.

    The auth-token is a JWT compliant token as the examples found here:
    https://github.com/eclipse-kuksa/kuksa-databroker/tree/main/certificates/jwt

    See also https://github.com/eclipse-kuksa/kuksa-databroker/blob/main/doc/authorization.md#jwt-access-token

    Upon reception of auth-token, server shall generate an auth-uuid in metadata
    that the client can use instead of auth-token in subsequent calls.

    """

    @staticmethod
    def Get(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/kuksa.val.v1.VAL/Get',
            kuksa_dot_val_dot_v1_dot_val__pb2.GetRequest.SerializeToString,
            kuksa_dot_val_dot_v1_dot_val__pb2.GetResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Set(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/kuksa.val.v1.VAL/Set',
            kuksa_dot_val_dot_v1_dot_val__pb2.SetRequest.SerializeToString,
            kuksa_dot_val_dot_v1_dot_val__pb2.SetResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StreamedUpdate(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/kuksa.val.v1.VAL/StreamedUpdate',
            kuksa_dot_val_dot_v1_dot_val__pb2.StreamedUpdateRequest.SerializeToString,
            kuksa_dot_val_dot_v1_dot_val__pb2.StreamedUpdateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Subscribe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/kuksa.val.v1.VAL/Subscribe',
            kuksa_dot_val_dot_v1_dot_val__pb2.SubscribeRequest.SerializeToString,
            kuksa_dot_val_dot_v1_dot_val__pb2.SubscribeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetServerInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/kuksa.val.v1.VAL/GetServerInfo',
            kuksa_dot_val_dot_v1_dot_val__pb2.GetServerInfoRequest.SerializeToString,
            kuksa_dot_val_dot_v1_dot_val__pb2.GetServerInfoResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: kuksa/val/v2/types.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'kuksa/val/v2/types.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18kuksa/val/v2/types.proto\x12\x0ckuksa.val.v2\x1a\x1fgoogle/protobuf/timestamp.proto\"^\n\tDatapoint\x12-\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\"\n\x05value\x18\x02 \x01(\x0b\x32\x13.kuksa.val.v2.Value\"\xaf\x04\n\x05Value\x12\x10\n\x06string\x18\x0b \x01(\tH\x00\x12\x0e\n\x04\x62ool\x18\x0c \x01(\x08H\x00\x12\x0f\n\x05int32\x18\r \x01(\x11H\x00\x12\x0f\n\x05int64\x18\x0e \x01(\x12H\x00\x12\x10\n\x06uint32\x18\x0f \x01(\rH\x00\x12\x10\n\x06uint64\x18\x10 \x01(\x04H\x00\x12\x0f\n\x05\x66loat\x18\x11 \x01(\x02H\x00\x12\x10\n\x06\x64ouble\x18\x12 \x01(\x01H\x00\x12\x31\n\x0cstring_array\x18\x15 \x01(\x0b\x32\x19.kuksa.val.v2.StringArrayH\x00\x12-\n\nbool_array\x18\x16 \x01(\x0b\x32\x17.kuksa.val.v2.BoolArrayH\x00\x12/\n\x0bint32_array\x18\x17 \x01(\x0b\x32\x18.kuksa.val.v2.Int32ArrayH\x00\x12/\n\x0bint64_array\x18\x18 \x01(\x0b\x32\x18.kuksa.val.v2.Int64ArrayH\x00\x12\x31\n\x0cuint32_array\x18\x19 \x01(\x0b\x32\x19.kuksa.val.v2.Uint32ArrayH\x00\x12\x31\n\x0cuint64_array\x18\x1a \x01(\x0b\x32\x19.kuksa.val.v2.Uint64ArrayH\x00\x12/\n\x0b\x66loat_array\x18\x1b \x01(\x0b\x32\x18.kuksa.val.v2.FloatArrayH\x00\x12\x31\n\x0c\x64ouble_array\x18\x1c \x01(\x0b\x32\x19.kuksa.val.v2.DoubleArrayH\x00\x42\r\n\x0btyped_value\"2\n\x08SignalID\x12\x0c\n\x02id\x18\x01 \x01(\x05H\x00\x12\x0e\n\x04path\x18\x02 \x01(\tH\x00\x42\x08\n\x06signal\"?\n\x05\x45rror\x12%\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x17.kuksa.val.v2.ErrorCode\x12\x0f\n\x07message\x18\x02 \x01(\t\"\xb6\x02\n\x08Metadata\x12\x0c\n\x04path\x18\t \x01(\t\x12\n\n\x02id\x18\n \x01(\x05\x12)\n\tdata_type\x18\x0b \x01(\x0e\x32\x16.kuksa.val.v2.DataType\x12+\n\nentry_type\x18\x0c \x01(\x0e\x32\x17.kuksa.val.v2.EntryType\x12\x13\n\x0b\x64\x65scription\x18\r \x01(\t\x12\x0f\n\x07\x63omment\x18\x0e \x01(\t\x12\x13\n\x0b\x64\x65precation\x18\x0f \x01(\t\x12\x0c\n\x04unit\x18\x10 \x01(\t\x12+\n\x0e\x61llowed_values\x18\x11 \x01(\x0b\x32\x13.kuksa.val.v2.Value\x12 \n\x03min\x18\x12 \x01(\x0b\x32\x13.kuksa.val.v2.Value\x12 \n\x03max\x18\x13 \x01(\x0b\x32\x13.kuksa.val.v2.Value\"\x1d\n\x0bStringArray\x12\x0e\n\x06values\x18\x01 \x03(\t\"\x1b\n\tBoolArray\x12\x0e\n\x06values\x18\x01 \x03(\x08\"\x1c\n\nInt32Array\x12\x0e\n\x06values\x18\x01 \x03(\x11\"\x1c\n\nInt64Array\x12\x0e\n\x06values\x18\x01 \x03(\x12\"\x1d\n\x0bUint32Array\x12\x0e\n\x06values\x18\x01 \x03(\r\"\x1d\n\x0bUint64Array\x12\x0e\n\x06values\x18\x01 \x03(\x04\"\x1c\n\nFloatArray\x12\x0e\n\x06values\x18\x01 \x03(\x02\"\x1d\n\x0b\x44oubleArray\x12\x0e\n\x06values\x18\x01 \x03(\x01*\x97\x01\n\tErrorCode\x12\x1a\n\x16\x45RROR_CODE_UNSPECIFIED\x10\x00\x12\x11\n\rERROR_CODE_OK\x10\x01\x12\x1f\n\x1b\x45RROR_CODE_INVALID_ARGUMENT\x10\x02\x12\x18\n\x14\x45RROR_CODE_NOT_FOUND\x10\x03\x12 \n\x1c\x45RROR_CODE_PERMISSION_DENIED\x10\x04*\xa9\x05\n\x08\x44\x61taType\x12\x19\n\x15\x44\x41TA_TYPE_UNSPECIFIED\x10\x00\x12\x14\n\x10\x44\x41TA_TYPE_STRING\x10\x01\x12\x15\n\x11\x44\x41TA_TYPE_BOOLEAN\x10\x02\x12\x12\n\x0e\x44\x41TA_TYPE_INT8\x10\x03\x12\x13\n\x0f\x44\x41TA_TYPE_INT16\x10\x04\x12\x13\n\x0f\x44\x41TA_TYPE_INT32\x10\x05\x12\x13\n\x0f\x44\x41TA_TYPE_INT64\x10\x06\x12\x13\n\x0f\x44\x41TA_TYPE_UINT8\x10\x07\x12\x14\n\x10\x44\x41TA_TYPE_UINT16\x10\x08\x12\x14\n\x10\x44\x41TA_TYPE_UINT32\x10\t\x12\x14\n\x10\x44\x41TA_TYPE_UINT64\x10\n\x12\x13\n\x0f\x44\x41TA_TYPE_FLOAT\x10\x0b\x12\x14\n\x10\x44\x41TA_TYPE_DOUBLE\x10\x0c\x12\x17\n\x13\x44\x41TA_TYPE_TIMESTAMP\x10\r\x12\x1a\n\x16\x44\x41TA_TYPE_STRING_ARRAY\x10\x14\x12\x1b\n\x17\x44\x41TA_TYPE_BOOLEAN_ARRAY\x10\x15\x12\x18\n\x14\x44\x41TA_TYPE_INT8_ARRAY\x10\x16\x12\x19\n\x15\x44\x41TA_TYPE_INT16_ARRAY\x10\x17\x12\x19\n\x15\x44\x41TA_TYPE_INT32_ARRAY\x10\x18\x12\x19\n\x15\x44\x41TA_TYPE_INT64_ARRAY\x10\x19\x12\x19\n\x15\x44\x41TA_TYPE_UINT8_ARRAY\x10\x1a\x12\x1a\n\x16\x44\x41TA_TYPE_UINT16_ARRAY\x10\x1b\x12\x1a\n\x16\x44\x41TA_TYPE_UINT32_ARRAY\x10\x1c\x12\x1a\n\x16\x44\x41TA_TYPE_UINT64_ARRAY\x10\x1d\x12\x19\n\x15\x44\x41TA_TYPE_FLOAT_ARRAY\x10\x1e\x12\x1a\n\x16\x44\x41TA_TYPE_DOUBLE_ARRAY\x10\x1f\x12\x1d\n\x19\x44\x41TA_TYPE_TIMESTAMP_ARRAY\x10 *q\n\tEntryType\x12\x1a\n\x16\x45NTRY_TYPE_UNSPECIFIED\x10\x00\x12\x18\n\x14\x45NTRY_TYPE_ATTRIBUTE\x10\x01\x12\x15\n\x11\x45NTRY_TYPE_SENSOR\x10\x02\x12\x17\n\x13\x45NTRY_TYPE_ACTUATOR\x10\x03\x42\x0eZ\x0ckuksa/val/v2b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kuksa.val.v2.types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\014kuksa/val/v2'
  _globals['_ERRORCODE']._serialized_start=1407
  _globals['_ERRORCODE']._serialized_end=1558
  _globals['_DATATYPE']._serialized_start=1561
  _globals['_DATATYPE']._serialized_end=2242
  _globals['_ENTRYTYPE']._serialized_start=2244
  _globals['_ENTRYTYPE']._serialized_end=2357
  _globals['_DATAPOINT']._serialized_start=75
  _globals['_DATAPOINT']._serialized_end=169
  _globals['_VALUE']._serialized_start=172
  _globals['_VALUE']._serialized_end=731
  _globals['_SIGNALID']._serialized_start=733
  _globals['_SIGNALID']._serialized_end=783
  _globals['_ERROR']._serialized_start=785
  _globals['_ERROR']._serialized_end=848
  _globals['_METADATA']._serialized_start=851
  _globals['_METADATA']._serialized_end=1161
  _globals['_STRINGARRAY']._serialized_start=1163
  _globals['_STRINGARRAY']._serialized_end=1192
  _globals['_BOOLARRAY']._serialized_start=1194
  _globals['_BOOLARRAY']._serialized_end=1221
  _globals['_INT32ARRAY']._serialized_start=1223
  _globals['_INT32ARRAY']._serialized_end=1251
  _globals['_INT64ARRAY']._serialized_start=1253
  _globals['_INT64ARRAY']._serialized_end=1281
  _globals['_UINT32ARRAY']._serialized_start=1283
  _globals['_UINT32ARRAY']._serialized_end=1312
  _globals['_UINT64ARRAY']._serialized_start=1314
  _globals['_UINT64ARRAY']._serialized_end=1343
  _globals['_FLOATARRAY']._serialized_start=1345
  _globals['_FLOATARRAY']._serialized_end=1373
  _globals['_DOUBLEARRAY']._serialized_start=1375
  _globals['_DOUBLEARRAY']._serialized_end=1404
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: sdv/databroker/v1/types.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'sdv/databroker/v1/types.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dsdv/databroker/v1/types.proto\x12\x11sdv.databroker.v1\x1a\x1fgoogle/protobuf/timestamp.proto\"\x1d\n\x0bStringArray\x12\x0e\n\x06values\x18\x01 \x03(\t\"\x1b\n\tBoolArray\x12\x0e\n\x06values\x18\x01 \x03(\x08\"\x1c\n\nInt32Array\x12\x0e\n\x06values\x18\x01 \x03(\x11\"\x1c\n\nInt64Array\x12\x0e\n\x06values\x18\x01 \x03(\x12\"\x1d\n\x0bUint32Array\x12\x0e\n\x06values\x18\x01 \x03(\r\"\x1d\n\x0bUint64Array\x12\x0e\n\x06values\x18\x01 \x03(\x04\"\x1c\n\nFloatArray\x12\x0e\n\x06values\x18\x01 \x03(\x02\"\x1d\n\x0b\x44oubleArray\x12\x0e\n\x06values\x18\x01 \x03(\x01\"\xe2\x06\n\tDatapoint\x12-\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12=\n\rfailure_value\x18\n \x01(\x0e\x32$.sdv.databroker.v1.Datapoint.FailureH\x00\x12\x16\n\x0cstring_value\x18\x0b \x01(\tH\x00\x12\x14\n\nbool_value\x18\x0c \x01(\x08H\x00\x12\x15\n\x0bint32_value\x18\r \x01(\x11H\x00\x12\x15\n\x0bint64_value\x18\x0e \x01(\x12H\x00\x12\x16\n\x0cuint32_value\x18\x0f \x01(\rH\x00\x12\x16\n\x0cuint64_value\x18\x10 \x01(\x04H\x00\x12\x15\n\x0b\x66loat_value\x18\x11 \x01(\x02H\x00\x12\x16\n\x0c\x64ouble_value\x18\x12 \x01(\x01H\x00\x12\x36\n\x0cstring_array\x18\x15 \x01(\x0b\x32\x1e.sdv.databroker.v1.StringArrayH\x00\x12\x32\n\nbool_array\x18\x16 \x01(\x0b\x32\x1c.sdv.databroker.v1.BoolArrayH\x00\x12\x34\n\x0bint32_array\x18\x17 \x01(\x0b\x32\x1d.sdv.databroker.v1.Int32ArrayH\x00\x12\x34\n\x0bint64_array\x18\x18 \x01(\x0b\x32\x1d.sdv.databroker.v1.Int64ArrayH\x00\x12\x36\n\x0cuint32_array\x18\x19 \x01(\x0b\x32\x1e.sdv.databroker.v1.Uint32ArrayH\x00\x12\x36\n\x0cuint64_array\x18\x1a \x01(\x0b\x32\x1e.sdv.databroker.v1.Uint64ArrayH\x00\x12\x34\n\x0b\x66loat_array\x18\x1b \x01(\x0b\x32\x1d.sdv.databroker.v1.FloatArrayH\x00\x12\x36\n\x0c\x64ouble_array\x18\x1c \x01(\x0b\x32\x1e.sdv.databroker.v1.DoubleArrayH\x00\"m\n\x07\x46\x61ilure\x12\x11\n\rINVALID_VALUE\x10\x00\x12\x11\n\rNOT_AVAILABLE\x10\x01\x12\x15\n\x11UNKNOWN_DATAPOINT\x10\x02\x12\x11\n\rACCESS_DENIED\x10\x03\x12\x12\n\x0eINTERNAL_ERROR\x10\x04\x42\x07\n\x05value\"\xe0\x02\n\x08Metadata\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x30\n\nentry_type\x18\x02 \x01(\x0e\x32\x1c.sdv.databroker.v1.EntryType\x12\x0c\n\x04name\x18\x04 \x01(\t\x12.\n\tdata_type\x18\x05 \x01(\x0e\x32\x1b.sdv.databroker.v1.DataType\x12\x32\n\x0b\x63hange_type\x18\x06 \x01(\x0e\x32\x1d.sdv.databroker.v1.ChangeType\x12\x13\n\x0b\x64\x65scription\x18\x07 \x01(\t\x12+\n\x07\x61llowed\x18\n \x01(\x0b\x32\x1a.sdv.databroker.v1.Allowed\x12\x30\n\x03min\x18\x0b \x01(\x0b\x32#.sdv.databroker.v1.ValueRestriction\x12\x30\n\x03max\x18\x0c \x01(\x0b\x32#.sdv.databroker.v1.ValueRestriction\"\x9c\x03\n\x07\x41llowed\x12\x37\n\rstring_values\x18\x01 \x01(\x0b\x32\x1e.sdv.databroker.v1.StringArrayH\x00\x12\x35\n\x0cint32_values\x18\x03 \x01(\x0b\x32\x1d.sdv.databroker.v1.Int32ArrayH\x00\x12\x35\n\x0cint64_values\x18\x04 \x01(\x0b\x32\x1d.sdv.databroker.v1.Int64ArrayH\x00\x12\x37\n\ruint32_values\x18\x05 \x01(\x0b\x32\x1e.sdv.databroker.v1.Uint32ArrayH\x00\x12\x37\n\ruint64_values\x18\x06 \x01(\x0b\x32\x1e.sdv.databroker.v1.Uint64ArrayH\x00\x12\x35\n\x0c\x66loat_values\x18\x07 \x01(\x0b\x32\x1d.sdv.databroker.v1.FloatArrayH\x00\x12\x37\n\rdouble_values\x18\x08 \x01(\x0b\x32\x1e.sdv.databroker.v1.DoubleArrayH\x00\x42\x08\n\x06values\"\xac\x01\n\x10ValueRestriction\x12\x10\n\x06string\x18\x01 \x01(\tH\x00\x12\x0e\n\x04\x62ool\x18\x02 \x01(\x08H\x00\x12\x0f\n\x05int32\x18\x03 \x01(\x11H\x00\x12\x0f\n\x05int64\x18\x04 \x01(\x12H\x00\x12\x10\n\x06uint32\x18\x05 \x01(\rH\x00\x12\x10\n\x06uint64\x18\x06 \x01(\x04H\x00\x12\x0f\n\x05\x66loat\x18\x07 \x01(\x02H\x00\x12\x10\n\x06\x64ouble\x18\x08 \x01(\x01H\x00\x42\r\n\x0btyped_value*\xe0\x02\n\x08\x44\x61taType\x12\n\n\x06STRING\x10\x00\x12\x08\n\x04\x42OOL\x10\x01\x12\x08\n\x04INT8\x10\x02\x12\t\n\x05INT16\x10\x03\x12\t\n\x05INT32\x10\x04\x12\t\n\x05INT64\x10\x05\x12\t\n\x05UINT8\x10\x06\x12\n\n\x06UINT16\x10\x07\x12\n\n\x06UINT32\x10\x08\x12\n\n\x06UINT64\x10\t\x12\t\n\x05\x46LOAT\x10\n\x12\n\n\x06\x44OUBLE\x10\x0b\x12\x10\n\x0cSTRING_ARRAY\x10\x14\x12\x0e\n\nBOOL_ARRAY\x10\x15\x12\x0e\n\nINT8_ARRAY\x10\x16\x12\x0f\n\x0bINT16_ARRAY\x10\x17\x12\x0f\n\x0bINT32_ARRAY\x10\x18\x12\x0f\n\x0bINT64_ARRAY\x10\x19\x12\x0f\n\x0bUINT8_ARRAY\x10\x1a\x12\x10\n\x0cUINT16_ARRAY\x10\x1b\x12\x10\n\x0cUINT32_ARRAY\x10\x1c\x12\x10\n\x0cUINT64_ARRAY\x10\x1d\x12\x0f\n\x0b\x46LOAT_ARRAY\x10\x1e\x12\x10\n\x0c\x44OUBLE_ARRAY\x10\x1f*s\n\x0e\x44\x61tapointError\x12\x15\n\x11UNKNOWN_DATAPOINT\x10\x00\x12\x10\n\x0cINVALID_TYPE\x10\x01\x12\x11\n\rACCESS_DENIED\x10\x02\x12\x12\n\x0eINTERNAL_ERROR\x10\x03\x12\x11\n\rOUT_OF_BOUNDS\x10\x04*q\n\tEntryType\x12\x1a\n\x16\x45NTRY_TYPE_UNSPECIFIED\x10\x00\x12\x15\n\x11\x45NTRY_TYPE_SENSOR\x10\x01\x12\x17\n\x13\x45NTRY_TYPE_ACTUATOR\x10\x02\x12\x18\n\x14\x45NTRY_TYPE_ATTRIBUTE\x10\x03*7\n\nChangeType\x12\n\n\x06STATIC\x10\x00\x12\r\n\tON_CHANGE\x10\x01\x12\x0e\n\nCONTINUOUS\x10\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sdv.databroker.v1.types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_DATATYPE']._serialized_start=2143
  _globals['_DATATYPE']._serialized_end=2495
  _globals['_DATAPOINTERROR']._serialized_start=2497
  _globals['_DATAPOINTERROR']._serialized_end=2612
  _globals['_ENTRYTYPE']._serialized_start=2614
  _globals['_ENTRYTYPE']._serialized_end=2727
  _globals['_CHANGETYPE']._serialized_start=2729
  _globals['_CHANGETYPE']._serialized_end=2784
  _globals['_STRINGARRAY']._serialized_start=85
  _globals['_STRINGARRAY']._serialized_end=114
  _globals['_BOOLARRAY']._serialized_start=116
  _globals['_BOOLARRAY']._serialized_end=143
  _globals['_INT32ARRAY']._serialized_start=145
  _globals['_INT32ARRAY']._serialized_end=173
  _globals['_INT64ARRAY']._serialized_start=175
  _globals['_INT64ARRAY']._serialized_end=203
  _globals['_UINT32ARRAY']._serialized_start=205
  _globals['_UINT32ARRAY']._serialized_end=234
  _globals['_UINT64ARRAY']._serialized_start=236
  _globals['_UINT64ARRAY']._serialized_end=265
  _globals['_FLOATARRAY']._serialized_start=267
  _globals['_FLOATARRAY']._serialized_end=295
  _globals['_DOUBLEARRAY']._serialized_start=297
  _globals['_DOUBLEARRAY']._serialized_end=326
  _globals['_DATAPOINT']._serialized_start=329
  _globals['_DATAPOINT']._serialized_end=1195
  _globals['_DATAPOINT_FAILURE']._serialized_start=1077
  _globals['_DATAPOINT_FAILURE']._serialized_end=1186
  _globals['_METADATA']._serialized_start=1198
  _globals['_METADATA']._serialized_end=1550
  _globals['_ALLOWED']._serialized_start=1553
  _globals['_ALLOWED']._serialized_end=1965
  _globals['_VALUERESTRICTION']._serialized_start=1968
  _globals['_VALUERESTRICTION']._serialized_end=2140
# @@protoc_insertion_point(module_scope)



########################################################################
# Copyright (c) 2022-2025 Robert Bosch GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
########################################################################

import asyncio
import datetime
import uuid

from google.protobuf import json_format
from google.protobuf import timestamp_pb2
import grpc
import grpc.aio
import pytest

from typing import Dict

from kuksa.val.v1 import types_pb2 as types_v1

from kuksa.val.v1 import val_pb2 as val_v1

# from kuksa.val.v1 import val_pb2_grpc as val_grpc_v1
from kuksa.val.v2 import types_pb2 as types_v2
from kuksa.val.v2 import val_pb2 as val_v2

# from kuksa.val.v2 import val_pb2_grpc as val_grpc_v2

import kuksa_client.grpc
from kuksa_client.grpc import Datapoint
from kuksa_client.grpc import DataEntry
from kuksa_client.grpc import DataType
from kuksa_client.grpc import EntryRequest
from kuksa_client.grpc import EntryType
from kuksa_client.grpc import EntryUpdate
from kuksa_client.grpc import Field
from kuksa_client.grpc import Metadata
from kuksa_client.grpc import MetadataField
from kuksa_client.grpc import ServerInfo
from kuksa_client.grpc import SubscribeEntry
from kuksa_client.grpc import ValueRestriction
from kuksa_client.grpc import View
from kuksa_client.grpc import VSSClientError
from kuksa_client.grpc.aio import VSSClient
from kuksa_client.grpc.aio import SubscriberManager


def generate_error(code: grpc.StatusCode, details: str):
    def set_error_context(_request, context):
        context.set_code(code)
        context.set_details(details)
    return set_error_context


class TestVSSClientError:
    def test_from_grpc_error(self):
        grpc_error = grpc.aio.AioRpcError(
            code=grpc.StatusCode.INVALID_ARGUMENT,
            initial_metadata=grpc.aio.Metadata(),
            trailing_metadata=grpc.aio.Metadata(),
            details='No datapoints requested',
        )
        client_error = VSSClientError.from_grpc_error(grpc_error)
        expected_client_error = VSSClientError(
            error={'code': 3, 'reason': 'invalid argument', 'message': 'No datapoints requested'}, errors=[],
        )
        assert client_error.error == expected_client_error.error
        assert client_error.errors == expected_client_error.errors

    def test_to_dict(self):
        error = types_v1.Error(
            code=404, reason="not_found", message="Does.Not.Exist not found"
        )
        errors = (types_v1.DataEntryError(path="Does.Not.Exist", error=error),)
        error = json_format.MessageToDict(
            error, preserving_proto_field_name=True)
        errors = [json_format.MessageToDict(
            err, preserving_proto_field_name=True) for err in errors]

        assert VSSClientError(error, errors).to_dict() == {
            'error': {'code': 404, 'reason': 'not_found', 'message': "Does.Not.Exist not found"},
            'errors': [{'path': 'Does.Not.Exist', 'error': {
                'code': 404, 'reason': 'not_found', 'message': "Does.Not.Exist not found",
            }}],
        }


class TestMetadata:
    def test_to_message_empty(self):
        assert Metadata().to_message() == types_v1.Metadata()

    def test_to_message_value_restriction_without_value_type(self):
        with pytest.raises(ValueError) as exc_info:
            assert (
                Metadata(value_restriction=ValueRestriction()).to_message()
                == types_v1.Metadata()
            )
        assert exc_info.value.args == (
            "Cannot set value_restriction from data type UNSPECIFIED",)

    @pytest.mark.parametrize('value_type', (
        DataType.INT8,
        DataType.INT16,
        DataType.INT32,
        DataType.INT64,
        DataType.INT8_ARRAY,
        DataType.INT16_ARRAY,
        DataType.INT32_ARRAY,
        DataType.INT64_ARRAY,
    ))
    @pytest.mark.parametrize('min_value, max_value, allowed_values', [
        (None, None, None),
        (None, None, []),
        (0, None, None),
        (-42, None, None),
        (None, 0, None),
        (None, 42, None),
        (None, None, [-42, 0, 42]),
        (-42, 42, [-42, 0, 42]),
    ])
    def test_to_from_message_signed_value_restriction(self, value_type, min_value, max_value, allowed_values):

        input_metadata = Metadata(value_restriction=ValueRestriction(
                            min=min_value, max=max_value, allowed_values=allowed_values,
                            ))

        if not allowed_values:
            # Empty array treated as None by KUKSA Metadata class
            allowed_values = None

        if (min_value, max_value, allowed_values) == (None, None, None):
            expected_message = types_v1.Metadata()
            output_metadata = Metadata()
        else:
            expected_message = types_v1.Metadata(
                value_restriction=types_v1.ValueRestriction(
                    signed=types_v1.ValueRestrictionInt(
                        min=min_value, max=max_value, allowed_values=allowed_values
                    ),
                )
            )
            output_metadata = Metadata(value_restriction=ValueRestriction(
                                min=min_value, max=max_value, allowed_values=allowed_values,
                                ))
        assert input_metadata.to_message(value_type) == expected_message
        assert Metadata.from_message(expected_message) == output_metadata

    @pytest.mark.parametrize('value_type', (
        DataType.UINT8,
        DataType.UINT16,
        DataType.UINT32,
        DataType.UINT64,
        DataType.UINT8_ARRAY,
        DataType.UINT16_ARRAY,
        DataType.UINT32_ARRAY,
        DataType.UINT64_ARRAY,
    ))
    @pytest.mark.parametrize('min_value, max_value, allowed_values', [
        (None, None, None),
        (None, None, []),
        (0, None, None),
        (None, 0, None),
        (None, 42, None),
        (None, None, [0, 12, 42]),
        (0, 42, [0, 12, 42]),
    ])
    def test_to_from_message_unsigned_value_restriction(self, value_type, min_value, max_value, allowed_values):

        input_metadata = Metadata(value_restriction=ValueRestriction(
                            min=min_value, max=max_value, allowed_values=allowed_values,
                            ))

        if not allowed_values:
            # Empty array treated as None by KUKSA Metadata class
            allowed_values = None

        if (min_value, max_value, allowed_values) == (None, None, None):
            expected_message = types_v1.Metadata()
            output_metadata = Metadata()
        else:
            expected_message = types_v1.Metadata(
                value_restriction=types_v1.ValueRestriction(
                    unsigned=types_v1.ValueRestrictionUint(
                        min=min_value, max=max_value, allowed_values=allowed_values
                    ),
                )
            )
            output_metadata = Metadata(value_restriction=ValueRestriction(
                                min=min_value, max=max_value, allowed_values=allowed_values,
                                ))

        assert input_metadata.to_message(value_type) == expected_message
        assert Metadata.from_message(expected_message) == output_metadata

    @pytest.mark.parametrize('value_type', (
        DataType.FLOAT,
        DataType.DOUBLE,
        DataType.FLOAT_ARRAY,
        DataType.DOUBLE_ARRAY,
    ))
    @pytest.mark.parametrize('min_value, max_value, allowed_values', [
        (None, None, None),
        (None, None, []),
        (0, None, None),
        (0.5, None, None),
        (None, 0, None),
        (None, 41.5, None),
        (None, None, [0.5, 12., 41.5]),
        (0.5, 41.5, [0.5, 12., 41.5]),
    ])
    def test_to_from_message_float_value_restriction(self, value_type, min_value, max_value, allowed_values):

        input_metadata = Metadata(value_restriction=ValueRestriction(
                            min=min_value, max=max_value, allowed_values=allowed_values,
                            ))

        if not allowed_values:
            # Empty array treated as None by KUKSA Metadata class
            allowed_values = None

        if (min_value, max_value, allowed_values) == (None, None, None):
            expected_message = types_v1.Metadata()
            output_metadata = Metadata()
        else:
            expected_message = types_v1.Metadata(
                value_restriction=types_v1.ValueRestriction(
                    floating_point=types_v1.ValueRestrictionFloat(
                        min=min_value, max=max_value, allowed_values=allowed_values
                    ),
                )
            )
            output_metadata = Metadata(value_restriction=ValueRestriction(
                                min=min_value, max=max_value, allowed_values=allowed_values,
                                ))

        assert input_metadata.to_message(value_type) == expected_message
        assert Metadata.from_message(expected_message) == output_metadata

    @pytest.mark.parametrize('value_type', (DataType.STRING, DataType.STRING_ARRAY))
    @pytest.mark.parametrize('allowed_values', [None, [], ['Hello', 'world']])
    def test_to_from_message_string_value_restriction(self, value_type, allowed_values):

        input_metadata = Metadata(value_restriction=ValueRestriction(
                            allowed_values=allowed_values,
                            ))

        if not allowed_values:
            # Empty array treated as None by KUKSA Metadata class
            allowed_values = None

        if allowed_values is None:
            expected_message = types_v1.Metadata()
            output_metadata = Metadata()
        else:
            expected_message = types_v1.Metadata(
                value_restriction=types_v1.ValueRestriction(
                    string=types_v1.ValueRestrictionString(
                        allowed_values=allowed_values
                    ),
                )
            )
            output_metadata = Metadata(value_restriction=ValueRestriction(
                            allowed_values=allowed_values,
                            ))

        assert input_metadata.to_message(value_type) == expected_message
        assert Metadata.from_message(expected_message) == output_metadata

    def test_metadata_from_message_value_restriction_no_type(self):
        """
        This intends to cover the case when the proto message has a value restriction, but
        no contents (type not specified)
        """
        input_message = types_v1.Metadata(value_restriction=types_v1.ValueRestriction())
        expected_metadata = Metadata()
        assert Metadata.from_message(input_message) == expected_metadata

    @pytest.mark.parametrize('metadata_dict, init_kwargs', [
        ({}, {}),
        ({'entry_type': 1}, {'entry_type': EntryType.ATTRIBUTE}),
        ({'entry_type': EntryType.ATTRIBUTE},
         {'entry_type': EntryType.ATTRIBUTE}),
        ({'entry_type': 'ATTRIBUTE'}, {'entry_type': EntryType.ATTRIBUTE}),
        ({'data_type': 1}, {'data_type': DataType.STRING}),
        ({'data_type': DataType.STRING}, {'data_type': DataType.STRING}),
        ({'data_type': 'STRING'}, {'data_type': DataType.STRING}),
        ({'description': 'a real description'},
         {'description': 'a real description'}),
        ({'comment': 'a real comment'}, {'comment': 'a real comment'}),
        ({'deprecation': 'a real deprecation'},
         {'deprecation': 'a real deprecation'}),
        ({'unit': 'a real unit'}, {'unit': 'a real unit'}),
        ({'value_restriction': {}}, {'value_restriction': ValueRestriction()}),
        ({'value_restriction': {'min': 12}}, {
         'value_restriction': ValueRestriction(min=12)}),
        ({'value_restriction': {'max': 42}}, {
         'value_restriction': ValueRestriction(max=42)}),
        ({'value_restriction': {'allowed_values': [-42, 42]}}, {
            'value_restriction': ValueRestriction(allowed_values=[-42, 42]),
        }),
        ({'value_restriction': {'min': -42, 'max': 42, 'allowed_values': [-42, 42]}}, {
            'value_restriction': ValueRestriction(min=-42, max=42, allowed_values=[-42, 42]),
        }),
    ])
    def test_from_dict(self, metadata_dict, init_kwargs):
        assert Metadata.from_dict(metadata_dict) == Metadata(**init_kwargs)

    @pytest.mark.parametrize('init_kwargs, metadata_dict', [
        ({}, {'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED'}),
        ({'entry_type': EntryType.ATTRIBUTE}, {
         'data_type': 'UNSPECIFIED', 'entry_type': 'ATTRIBUTE'}),
        ({'data_type': DataType.STRING}, {
         'data_type': 'STRING', 'entry_type': 'UNSPECIFIED'}),
        ({'description': 'a real description'}, {
            'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED', 'description': 'a real description',
        }),
        ({'comment': 'a real comment'}, {
            'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED', 'comment': 'a real comment',
        }),
        ({'deprecation': 'a real deprecation'}, {
            'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED', 'deprecation': 'a real deprecation',
        }),
        ({'unit': 'a real unit'}, {'data_type': 'UNSPECIFIED',
         'entry_type': 'UNSPECIFIED', 'unit': 'a real unit'}),
        ({'value_restriction': ValueRestriction()}, {
            'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED', 'value_restriction': {},
        }),
        ({'value_restriction': ValueRestriction(min=12)}, {
            'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED',
            'value_restriction': {'min': 12},
        }),
        ({'value_restriction': ValueRestriction(max=42)}, {
            'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED',
            'value_restriction': {'max': 42},
        }),
        ({'value_restriction': ValueRestriction(allowed_values=[-42, 42])}, {
            'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED',
            'value_restriction': {'allowed_values': [-42, 42]},
        }),
        ({'value_restriction': ValueRestriction(min=-42, max=42, allowed_values=[-42, 42])}, {
            'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED',
            'value_restriction': {'min': -42, 'max': 42, 'allowed_values': [-42, 42]},
        }),
    ])
    def test_to_dict(self, init_kwargs, metadata_dict):
        assert Metadata(**init_kwargs).to_dict() == metadata_dict


class TestDatapoint:

    @pytest.mark.parametrize(
        "value_type, init_args, message_v1, message_v2",
        [
            (DataType.BOOLEAN, (None,), types_v1.Datapoint(), types_v2.Datapoint()),
            (
                DataType.BOOLEAN,
                ("False",),
                types_v1.Datapoint(bool=False),
                types_v2.Datapoint(value=types_v2.Value(bool=False)),
            ),
            (
                DataType.BOOLEAN,
                ("false",),
                types_v1.Datapoint(bool=False),
                types_v2.Datapoint(value=types_v2.Value(bool=False)),
            ),
            (
                DataType.BOOLEAN,
                ("F",),
                types_v1.Datapoint(bool=False),
                types_v2.Datapoint(value=types_v2.Value(bool=False)),
            ),
            (
                DataType.BOOLEAN,
                ("f",),
                types_v1.Datapoint(bool=False),
                types_v2.Datapoint(value=types_v2.Value(bool=False)),
            ),
            (
                DataType.BOOLEAN,
                (True, datetime.datetime(2022, 11, 16, tzinfo=datetime.timezone.utc)),
                types_v1.Datapoint(
                    bool=True,
                    timestamp=timestamp_pb2.Timestamp(seconds=1668556800),
                ),
                types_v2.Datapoint(
                    value=types_v2.Value(bool=True),
                    timestamp=timestamp_pb2.Timestamp(seconds=1668556800),
                ),
            ),
            (
                DataType.INT8_ARRAY,
                ("[-128, 127]",),
                types_v1.Datapoint(int32_array=types_v1.Int32Array(values=[-128, 127])),
                types_v2.Datapoint(
                    value=types_v2.Value(
                        int32_array=types_v2.Int32Array(values=[-128, 127])
                    ),
                ),
            ),
        ],
    )
    def test_to_message(self, value_type, init_args, message_v1, message_v2):
        assert Datapoint(*init_args).v1_to_message(value_type) == message_v1
        assert Datapoint(*init_args).v2_to_message(value_type) == message_v2

    @pytest.mark.parametrize('value_type', [DataType.UNSPECIFIED, DataType.TIMESTAMP, DataType.TIMESTAMP_ARRAY])
    def test_v1_to_message_unsupported_value_type(self, value_type):
        with pytest.raises(ValueError) as exc_info:
            Datapoint(42).v1_to_message(value_type)
        assert exc_info.value.args[0].startswith(
            "Cannot determine which field to set with data type"
        )

    @pytest.mark.parametrize(
        "value_type",
        [DataType.UNSPECIFIED, DataType.TIMESTAMP, DataType.TIMESTAMP_ARRAY],
    )
    def test_v2_to_message_unsupported_value_type(self, value_type):
        with pytest.raises(ValueError) as exc_info:
            Datapoint(42).v2_to_message(value_type)
        assert exc_info.value.args[0].startswith(
            'Cannot determine which field to set with data type')

    @pytest.mark.parametrize('init_kwargs, datapoint_dict', [
        ({}, {}),
        ({'value': 42}, {'value': 42}),
        ({'timestamp': datetime.datetime(2022, 11, 16, tzinfo=datetime.timezone.utc)}, {
            'timestamp': '2022-11-16T00:00:00+00:00',
        }),
        ({'value': 42, 'timestamp': datetime.datetime(2022, 11, 16, tzinfo=datetime.timezone.utc)}, {
            'value': 42, 'timestamp': '2022-11-16T00:00:00+00:00',
        }),
    ])
    def test_to_dict(self, init_kwargs, datapoint_dict):
        assert Datapoint(**init_kwargs).to_dict() == datapoint_dict


class TestDataEntry:
    @pytest.mark.parametrize('value, actuator_target, metadata, entry_dict', [
        (None, None, None, {}),
        (Datapoint(42), None, None, {'value': {'value': 42}}),
        (None, Datapoint(42), None, {'actuator_target': {'value': 42}}),
        (None, None, Metadata(), {'metadata': {
         'data_type': 'UNSPECIFIED', 'entry_type': 'UNSPECIFIED'}}),
    ])
    def test_to_dict(self, value, actuator_target, metadata, entry_dict):
        entry_dict.update({'path': 'Test.Path'})
        assert DataEntry('Test.Path', value, actuator_target,
                         metadata).to_dict() == entry_dict


class TestEntryUpdate:
    @pytest.mark.parametrize('entry, fields, update_dict', [
        (DataEntry(path='Test.Path'), (), {
         'entry': {'path': 'Test.Path'}, 'fields': []}),
        (DataEntry(path='Test.Path', value=Datapoint(42)), (Field.VALUE,), {
            'entry': {'path': 'Test.Path', 'value': {'value': 42}}, 'fields': ['VALUE'],
        }),
    ])
    def test_to_dict(self, entry, fields, update_dict):
        assert EntryUpdate(entry, fields).to_dict() == update_dict


@pytest.mark.asyncio
class TestVSSClient:

    @pytest.mark.usefixtures("secure_mocked_databroker")
    async def test_secure_connection(
        self, unused_tcp_port, resources_path, val_servicer_v1
    ):
        val_servicer_v1.GetServerInfo.return_value = val_v1.GetServerInfoResponse(
            name="test_server", version="1.2.3"
        )
        async with VSSClient('localhost', unused_tcp_port,
                             root_certificates=resources_path / 'test-ca.pem',
                             ensure_startup_connection=True
                             ):
            assert val_servicer_v1.GetServerInfo.call_count == 1

    async def test_get_current_values(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection test

        mocker.patch.object(client, 'get', return_value=[
            DataEntry('Vehicle.Speed', value=Datapoint(
                42.0, datetime.datetime(
                    2022, 11, 7, 16, 18, 35, 247307, tzinfo=datetime.timezone.utc),
            )),
            DataEntry('Vehicle.ADAS.ABS.IsActive', value=Datapoint(True)),
            DataEntry('Vehicle.Chassis.Height', value=Datapoint(666)),
        ])
        assert await client.get_current_values([
            'Vehicle.Speed', 'Vehicle.ADAS.ABS.IsActive', 'Vehicle.Chassis.Height',
        ]) == {
            'Vehicle.Speed': Datapoint(
                42.0, datetime.datetime(
                    2022, 11, 7, 16, 18, 35, 247307, tzinfo=datetime.timezone.utc),
            ),
            'Vehicle.ADAS.ABS.IsActive': Datapoint(True),
            'Vehicle.Chassis.Height': Datapoint(666),
        }
        assert list(client.get.call_args_list[0][1]['entries']) == [
            EntryRequest('Vehicle.Speed', View.CURRENT_VALUE, (Field.VALUE,)),
            EntryRequest('Vehicle.ADAS.ABS.IsActive',
                         View.CURRENT_VALUE, (Field.VALUE,)),
            EntryRequest('Vehicle.Chassis.Height',
                         View.CURRENT_VALUE, (Field.VALUE,)),
        ]

    async def test_get_target_values(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection check
        mocker.patch.object(client, 'get', return_value=[
            DataEntry('Vehicle.ADAS.ABS.IsActive', actuator_target=Datapoint(
                True, datetime.datetime(
                    2022, 11, 7, tzinfo=datetime.timezone.utc),
            )),
            DataEntry('Vehicle.Chassis.SteeringWheel.Tilt',
                      actuator_target=Datapoint(42)),
        ])
        assert await client.get_target_values([
            'Vehicle.ADAS.ABS.IsActive', 'Vehicle.Chassis.SteeringWheel.Tilt',
        ]) == {
            'Vehicle.ADAS.ABS.IsActive': Datapoint(True, datetime.datetime(2022, 11, 7, tzinfo=datetime.timezone.utc)),
            'Vehicle.Chassis.SteeringWheel.Tilt': Datapoint(42),
        }
        assert list(client.get.call_args_list[0][1]['entries']) == [
            EntryRequest('Vehicle.ADAS.ABS.IsActive',
                         View.TARGET_VALUE, (Field.ACTUATOR_TARGET,)),
            EntryRequest('Vehicle.Chassis.SteeringWheel.Tilt',
                         View.TARGET_VALUE, (Field.ACTUATOR_TARGET,)),
        ]

    async def test_get_metadata(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection check
        mocker.patch.object(client, 'get', return_value=[
            DataEntry('Vehicle.Speed', metadata=Metadata(
                entry_type=EntryType.SENSOR)),
            DataEntry('Vehicle.ADAS.ABS.IsActive', metadata=Metadata(
                entry_type=EntryType.ACTUATOR)),
            DataEntry('Vehicle.Chassis.Height', metadata=Metadata(
                entry_type=EntryType.ATTRIBUTE)),
        ])
        assert await client.get_metadata([
            'Vehicle.Speed',
            'Vehicle.ADAS.ABS.IsActive',
            'Vehicle.Chassis.Height',
        ], field=MetadataField.ENTRY_TYPE) == {
            'Vehicle.Speed': Metadata(entry_type=EntryType.SENSOR),
            'Vehicle.ADAS.ABS.IsActive': Metadata(entry_type=EntryType.ACTUATOR),
            'Vehicle.Chassis.Height': Metadata(entry_type=EntryType.ATTRIBUTE),
        }
        assert list(client.get.call_args_list[0][1]['entries']) == [
            EntryRequest('Vehicle.Speed', View.METADATA,
                         (Field.METADATA_ENTRY_TYPE,)),
            EntryRequest('Vehicle.ADAS.ABS.IsActive',
                         View.METADATA, (Field.METADATA_ENTRY_TYPE,)),
            EntryRequest('Vehicle.Chassis.Height', View.METADATA,
                         (Field.METADATA_ENTRY_TYPE,)),
        ]

    async def test_set_current_values(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection check
        mocker.patch.object(client, 'set')
        await client.set_current_values({
            'Vehicle.Speed': Datapoint(42.0,
                                       datetime.datetime(2022, 11, 7, 16, 18, 35, 247307,
                                                         tzinfo=datetime.timezone.utc)),
            'Vehicle.ADAS.ABS.IsActive': Datapoint(True),
            'Vehicle.Chassis.Height': Datapoint(666),
        })
        assert list(client.set.call_args_list[0][1]['updates']) == [
            EntryUpdate(DataEntry('Vehicle.Speed', value=Datapoint(
                42.0, datetime.datetime(
                    2022, 11, 7, 16, 18, 35, 247307, tzinfo=datetime.timezone.utc),
            )), (Field.VALUE,)),
            EntryUpdate(DataEntry('Vehicle.ADAS.ABS.IsActive',
                        value=Datapoint(True)), (Field.VALUE,)),
            EntryUpdate(DataEntry('Vehicle.Chassis.Height',
                        value=Datapoint(666)), (Field.VALUE,)),
        ]

    async def test_set_target_values(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection check
        mocker.patch.object(client, 'set')
        await client.set_target_values({
            'Vehicle.ADAS.ABS.IsActive': Datapoint(True, datetime.datetime(2022, 11, 7, tzinfo=datetime.timezone.utc)),
            'Vehicle.Chassis.SteeringWheel.Tilt': Datapoint(42),
        })
        assert list(client.set.call_args_list[0][1]['updates']) == [
            EntryUpdate(DataEntry('Vehicle.ADAS.ABS.IsActive', actuator_target=Datapoint(
                True, datetime.datetime(
                    2022, 11, 7, tzinfo=datetime.timezone.utc),
            )), (Field.ACTUATOR_TARGET,)),
            EntryUpdate(DataEntry('Vehicle.Chassis.SteeringWheel.Tilt',
                        actuator_target=Datapoint(42)), (Field.ACTUATOR_TARGET,)),
        ]

    async def test_set_metadata(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection check
        mocker.patch.object(client, 'set')
        await client.set_metadata({
            'Vehicle.Speed': Metadata(entry_type=EntryType.SENSOR),
            'Vehicle.ADAS.ABS.IsActive': Metadata(entry_type=EntryType.ACTUATOR),
            'Vehicle.Chassis.Height': Metadata(entry_type=EntryType.ATTRIBUTE),
        }, field=MetadataField.ENTRY_TYPE)
        assert list(client.set.call_args_list[0][1]['updates']) == [
            EntryUpdate(DataEntry(
                'Vehicle.Speed', metadata=Metadata(entry_type=EntryType.SENSOR),
            ), (Field.METADATA_ENTRY_TYPE,)),
            EntryUpdate(DataEntry(
                'Vehicle.ADAS.ABS.IsActive', metadata=Metadata(entry_type=EntryType.ACTUATOR),
            ), (Field.METADATA_ENTRY_TYPE,)),
            EntryUpdate(DataEntry(
                'Vehicle.Chassis.Height', metadata=Metadata(entry_type=EntryType.ATTRIBUTE),
            ), (Field.METADATA_ENTRY_TYPE,)),
        ]

    async def test_subscribe_current_values(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection check

        async def subscribe_response_stream(**kwargs):
            yield [
                EntryUpdate(DataEntry('Vehicle.Speed', value=Datapoint(
                    42.0, datetime.datetime(
                        2022, 11, 7, 16, 18, 35, 247307, tzinfo=datetime.timezone.utc),
                )), (Field.VALUE,)),
                EntryUpdate(DataEntry('Vehicle.ADAS.ABS.IsActive',
                            value=Datapoint(True)), (Field.VALUE,)),
                EntryUpdate(DataEntry('Vehicle.Chassis.Height',
                            value=Datapoint(666)), (Field.VALUE,)),
            ]
        mocker.patch.object(client, 'v2_subscribe',
                            side_effect=subscribe_response_stream)

        received_updates = {}
        async for updates in client.subscribe_current_values([
            'Vehicle.Speed', 'Vehicle.ADAS.ABS.IsActive', 'Vehicle.Chassis.Height',
        ]):
            received_updates.update(updates)

        assert list(client.v2_subscribe.call_args_list[0][1]['paths']) == [
            'Vehicle.Speed',
            'Vehicle.ADAS.ABS.IsActive',
            'Vehicle.Chassis.Height',
        ]
        assert received_updates == {
            'Vehicle.Speed': Datapoint(42.0,
                                       datetime.datetime(2022, 11, 7, 16, 18, 35, 247307,
                                                         tzinfo=datetime.timezone.utc)),
            'Vehicle.ADAS.ABS.IsActive': Datapoint(True),
            'Vehicle.Chassis.Height': Datapoint(666),
        }

    async def test_subscribe_target_values(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection check

        async def subscribe_response_stream(**kwargs):
            yield [
                EntryUpdate(DataEntry('Vehicle.ADAS.ABS.IsActive', actuator_target=Datapoint(
                    True, datetime.datetime(
                        2022, 11, 7, tzinfo=datetime.timezone.utc),
                )), (Field.ACTUATOR_TARGET,)),
                EntryUpdate(DataEntry('Vehicle.Chassis.SteeringWheel.Tilt',
                            actuator_target=Datapoint(42)), (Field.ACTUATOR_TARGET,)),
            ]
        mocker.patch.object(client, 'v2_subscribe_actuation_requests',
                            side_effect=subscribe_response_stream)

        received_updates = {}
        async for updates in client.subscribe_target_values([
            'Vehicle.ADAS.ABS.IsActive', 'Vehicle.Chassis.SteeringWheel.Tilt',
        ]):
            received_updates.update(updates)

        assert list(client.v2_subscribe_actuation_requests.call_args_list[0][1]['paths']) == [
            'Vehicle.ADAS.ABS.IsActive',
            'Vehicle.Chassis.SteeringWheel.Tilt',
        ]
        assert received_updates == {
            'Vehicle.ADAS.ABS.IsActive': Datapoint(True, datetime.datetime(2022, 11, 7, tzinfo=datetime.timezone.utc)),
            'Vehicle.Chassis.SteeringWheel.Tilt': Datapoint(42),
        }

    async def test_subscribe_metadata(self, mocker, unused_tcp_port):
        client = VSSClient('127.0.0.1', unused_tcp_port)
        client.connected = True  # To bypass connection check

        async def subscribe_response_stream(**kwargs):
            yield [
                EntryUpdate(DataEntry(
                    'Vehicle.Speed', metadata=Metadata(entry_type=EntryType.SENSOR),
                ), (Field.METADATA_ENTRY_TYPE,)),
                EntryUpdate(DataEntry(
                    'Vehicle.ADAS.ABS.IsActive', metadata=Metadata(entry_type=EntryType.ACTUATOR),
                ), (Field.METADATA_ENTRY_TYPE,)),
                EntryUpdate(DataEntry(
                    'Vehicle.Chassis.Height', metadata=Metadata(entry_type=EntryType.ATTRIBUTE),
                ), (Field.METADATA_ENTRY_TYPE,)),
            ]
        mocker.patch.object(client, 'subscribe',
                            side_effect=subscribe_response_stream)

        received_updates = {}
        async for updates in client.subscribe_metadata([
            'Vehicle.Speed', 'Vehicle.ADAS.ABS.IsActive', 'Vehicle.Chassis.Height',
        ], field=MetadataField.ENTRY_TYPE):
            received_updates.update(updates)

        assert list(client.subscribe.call_args_list[0][1]['entries']) == [
            SubscribeEntry('Vehicle.Speed', View.METADATA,
                           (Field.METADATA_ENTRY_TYPE,)),
            SubscribeEntry('Vehicle.ADAS.ABS.IsActive',
                           View.METADATA, (Field.METADATA_ENTRY_TYPE,)),
            SubscribeEntry('Vehicle.Chassis.Height',
                           View.METADATA, (Field.METADATA_ENTRY_TYPE,)),
        ]
        assert received_updates == {
            'Vehicle.Speed': Metadata(entry_type=EntryType.SENSOR),
            'Vehicle.ADAS.ABS.IsActive': Metadata(entry_type=EntryType.ACTUATOR),
            'Vehicle.Chassis.Height': Metadata(entry_type=EntryType.ATTRIBUTE),
        }

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_get_some_entries(self, unused_tcp_port, val_servicer_v1):
        val_servicer_v1.Get.return_value = val_v1.GetResponse(
            entries=[
                types_v1.DataEntry(
                    path="Vehicle.Speed",
                    value=types_v1.Datapoint(
                        timestamp=timestamp_pb2.Timestamp(
                            seconds=1667837915, nanos=247307674
                        ),
                        float=42.0,
                    ),
                ),
                types_v1.DataEntry(
                    path="Vehicle.ADAS.ABS.IsActive",
                    actuator_target=types_v1.Datapoint(bool=True),
                ),
                types_v1.DataEntry(
                    path="Vehicle.Chassis.Height",
                    metadata=types_v1.Metadata(
                        data_type=types_v1.DATA_TYPE_UINT16,
                        entry_type=types_v1.ENTRY_TYPE_ATTRIBUTE,
                        description="Overall vehicle height, in mm.",
                        comment="No comment.",
                        deprecation="V2.1 moved to Vehicle.Height",
                        unit="mm",
                    ),
                ),
                types_v1.DataEntry(
                    path="Vehicle.Chassis.Height",
                    metadata=types_v1.Metadata(data_type=types_v1.DATA_TYPE_UINT16),
                ),
                types_v1.DataEntry(
                    path="Vehicle.Chassis.Height",
                    metadata=types_v1.Metadata(
                        entry_type=types_v1.ENTRY_TYPE_ATTRIBUTE
                    ),
                ),
                types_v1.DataEntry(
                    path="Vehicle.Chassis.Height",
                    metadata=types_v1.Metadata(
                        description="Overall vehicle height, in mm."
                    ),
                ),
                types_v1.DataEntry(
                    path="Vehicle.Chassis.Height",
                    metadata=types_v1.Metadata(comment="No comment."),
                ),
                types_v1.DataEntry(
                    path="Vehicle.Chassis.Height",
                    metadata=types_v1.Metadata(
                        deprecation="V2.1 moved to Vehicle.Height"
                    ),
                ),
                types_v1.DataEntry(
                    path="Vehicle.Chassis.Height", metadata=types_v1.Metadata(unit="mm")
                ),
                types_v1.DataEntry(
                    path="Vehicle.CurrentLocation.Heading",
                    metadata=types_v1.Metadata(
                        value_restriction=types_v1.ValueRestriction(
                            floating_point=types_v1.ValueRestrictionFloat(
                                min=0, max=360
                            ),
                        )
                    ),
                ),
                types_v1.DataEntry(
                    path="Dummy.With.Allowed.Values",
                    metadata=types_v1.Metadata(
                        value_restriction=types_v1.ValueRestriction(
                            signed=types_v1.ValueRestrictionInt(
                                allowed_values=[12, 42, 666]
                            ),
                        )
                    ),
                ),
            ]
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            entries = await client.get(entries=(entry for entry in (
                # generator is intentional as get accepts Iterable
                EntryRequest('Vehicle.Speed',
                             View.CURRENT_VALUE, (Field.VALUE,)),
                EntryRequest('Vehicle.ADAS.ABS.IsActive',
                             View.TARGET_VALUE, (Field.ACTUATOR_TARGET,)),
                EntryRequest('Vehicle.Chassis.Height',
                             View.METADATA, (Field.METADATA,)),
                EntryRequest('Vehicle.Chassis.Height',
                             View.METADATA, (Field.METADATA_DATA_TYPE,)),
                EntryRequest('Vehicle.Chassis.Height',
                             View.METADATA, (Field.METADATA_DESCRIPTION,)),
                EntryRequest('Vehicle.Chassis.Height',
                             View.METADATA, (Field.METADATA_ENTRY_TYPE,)),
                EntryRequest('Vehicle.Chassis.Height',
                             View.METADATA, (Field.METADATA_COMMENT,)),
                EntryRequest('Vehicle.Chassis.Height',
                             View.METADATA, (Field.METADATA_DEPRECATION,)),
                EntryRequest('Vehicle.Chassis.Height',
                             View.METADATA, (Field.METADATA_UNIT,)),
                EntryRequest('Vehicle.CurrentLocation.Heading',
                             View.METADATA, (Field.METADATA_VALUE_RESTRICTION,)),
                EntryRequest('Dummy.With.Allowed.Values', View.METADATA,
                             (Field.METADATA_VALUE_RESTRICTION,)),
            )))
            assert entries == [
                DataEntry('Vehicle.Speed', value=Datapoint(
                    42.0, datetime.datetime(
                        2022, 11, 7, 16, 18, 35, 247307, tzinfo=datetime.timezone.utc),
                )),
                DataEntry('Vehicle.ADAS.ABS.IsActive',
                          actuator_target=Datapoint(True)),
                DataEntry('Vehicle.Chassis.Height', metadata=Metadata(
                    data_type=DataType.UINT16,
                    entry_type=EntryType.ATTRIBUTE,
                    description="Overall vehicle height, in mm.",
                    comment="No comment.",
                    deprecation="V2.1 moved to Vehicle.Height",
                    unit="mm",
                )),
                DataEntry('Vehicle.Chassis.Height',
                          metadata=Metadata(data_type=DataType.UINT16)),
                DataEntry('Vehicle.Chassis.Height', metadata=Metadata(
                    entry_type=EntryType.ATTRIBUTE)),
                DataEntry('Vehicle.Chassis.Height', metadata=Metadata(
                    description="Overall vehicle height, in mm.")),
                DataEntry('Vehicle.Chassis.Height',
                          metadata=Metadata(comment="No comment.")),
                DataEntry('Vehicle.Chassis.Height', metadata=Metadata(
                    deprecation="V2.1 moved to Vehicle.Height")),
                DataEntry('Vehicle.Chassis.Height',
                          metadata=Metadata(unit="mm")),
                DataEntry('Vehicle.CurrentLocation.Heading', metadata=Metadata(
                    value_restriction=ValueRestriction(min=0, max=360),
                )),
                DataEntry('Dummy.With.Allowed.Values', metadata=Metadata(
                    value_restriction=ValueRestriction(
                        allowed_values=[12, 42, 666]),
                )),
            ]
            assert (
                val_servicer_v1.Get.call_args[0][0].entries
                == val_v1.GetRequest(
                    entries=(
                        val_v1.EntryRequest(
                            path="Vehicle.Speed",
                            view=types_v1.VIEW_CURRENT_VALUE,
                            fields=(types_v1.FIELD_VALUE,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.ADAS.ABS.IsActive",
                            view=types_v1.VIEW_TARGET_VALUE,
                            fields=(types_v1.FIELD_ACTUATOR_TARGET,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.Chassis.Height",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.Chassis.Height",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA_DATA_TYPE,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.Chassis.Height",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA_DESCRIPTION,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.Chassis.Height",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA_ENTRY_TYPE,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.Chassis.Height",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA_COMMENT,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.Chassis.Height",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA_DEPRECATION,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.Chassis.Height",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA_UNIT,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.CurrentLocation.Heading",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA_VALUE_RESTRICTION,),
                        ),
                        val_v1.EntryRequest(
                            path="Dummy.With.Allowed.Values",
                            view=types_v1.VIEW_METADATA,
                            fields=(types_v1.FIELD_METADATA_VALUE_RESTRICTION,),
                        ),
                    )
                ).entries
            )

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_get_no_entries_requested(self, unused_tcp_port, val_servicer_v1):
        val_servicer_v1.Get.side_effect = generate_error(
            grpc.StatusCode.INVALID_ARGUMENT, "No datapoints requested"
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            with pytest.raises(kuksa_client.grpc.VSSClientError) as exc_info:
                await client.get(entries=[])

            assert exc_info.value.args == kuksa_client.grpc.VSSClientError(error={
                'code': grpc.StatusCode.INVALID_ARGUMENT.value[0],
                'reason': grpc.StatusCode.INVALID_ARGUMENT.value[1],
                'message': 'No datapoints requested',
            }, errors=[]).args
            assert val_servicer_v1.Get.call_args[0][0] == val_v1.GetRequest()

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_get_unset_entries(self, unused_tcp_port, val_servicer_v1):
        val_servicer_v1.Get.return_value = val_v1.GetResponse(
            entries=[
                types_v1.DataEntry(path="Vehicle.Speed"),
                types_v1.DataEntry(path="Vehicle.ADAS.ABS.IsActive"),
            ]
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            entries = await client.get(entries=(
                EntryRequest('Vehicle.Speed',
                             View.CURRENT_VALUE, (Field.VALUE,)),
                EntryRequest('Vehicle.ADAS.ABS.IsActive',
                             View.TARGET_VALUE, (Field.ACTUATOR_TARGET,)),
            ))
        assert entries == [DataEntry('Vehicle.Speed'), DataEntry(
            'Vehicle.ADAS.ABS.IsActive')]

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_get_nonexistent_entries(self, unused_tcp_port, val_servicer_v1):
        error = types_v1.Error(
            code=404, reason="not_found", message="Does.Not.Exist not found"
        )
        errors = (types_v1.DataEntryError(path="Does.Not.Exist", error=error),)
        val_servicer_v1.Get.return_value = val_v1.GetResponse(
            error=error, errors=errors
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            with pytest.raises(VSSClientError):
                await client.get(entries=(
                    EntryRequest('Does.Not.Exist',
                                 View.CURRENT_VALUE, (Field.VALUE,)),
                ))

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_set_some_updates_v1(self, unused_tcp_port, val_servicer_v1):
        val_servicer_v1.Get.return_value = val_v1.GetResponse(
            entries=(
                types_v1.DataEntry(
                    path="Vehicle.Speed",
                    metadata=types_v1.Metadata(data_type=types_v1.DATA_TYPE_FLOAT),
                ),
                types_v1.DataEntry(
                    path="Vehicle.ADAS.ABS.IsActive",
                    metadata=types_v1.Metadata(data_type=types_v1.DATA_TYPE_BOOLEAN),
                ),
                types_v1.DataEntry(
                    path="Vehicle.Cabin.Door.Row1.Left.Shade.Position",
                    metadata=types_v1.Metadata(data_type=types_v1.DATA_TYPE_UINT8),
                ),
            )
        )
        val_servicer_v1.Set.return_value = val_v1.SetResponse()
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            await client.set(
                updates=[
                    EntryUpdate(
                        DataEntry("Vehicle.Speed", value=Datapoint(value=42.0)),
                        (Field.VALUE,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.ADAS.ABS.IsActive",
                            actuator_target=Datapoint(value=False),
                        ),
                        (Field.ACTUATOR_TARGET,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.ADAS.CruiseControl.Error",
                            metadata=Metadata(
                                data_type=DataType.BOOLEAN,
                                entry_type=EntryType.SENSOR,
                                description="Indicates if cruise control system incurred and error condition.",
                                comment="No comment",
                                deprecation="Never to be deprecated",
                                unit=None,
                                value_restriction=None,
                            ),
                        ),
                        (Field.METADATA,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.ADAS.CruiseControl.Error",
                            metadata=Metadata(
                                data_type=DataType.BOOLEAN,
                            ),
                        ),
                        (Field.METADATA_DATA_TYPE,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.ADAS.CruiseControl.Error",
                            metadata=Metadata(
                                description="Indicates if cruise control system incurred and error condition.",
                            ),
                        ),
                        (Field.METADATA_DESCRIPTION,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.ADAS.CruiseControl.Error",
                            metadata=Metadata(
                                entry_type=EntryType.SENSOR,
                            ),
                        ),
                        (Field.METADATA_ENTRY_TYPE,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.ADAS.CruiseControl.Error",
                            metadata=Metadata(
                                comment="No comment",
                            ),
                        ),
                        (Field.METADATA_COMMENT,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.ADAS.CruiseControl.Error",
                            metadata=Metadata(
                                deprecation="Never to be deprecated",
                            ),
                        ),
                        (Field.METADATA_DEPRECATION,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.Cabin.Door.Row1.Left.Shade.Position",
                            metadata=Metadata(
                                unit="percent",
                            ),
                        ),
                        (Field.METADATA_UNIT,),
                    ),
                    EntryUpdate(
                        DataEntry(
                            "Vehicle.Cabin.Door.Row1.Left.Shade.Position",
                            metadata=Metadata(
                                value_restriction=ValueRestriction(min=0, max=100),
                            ),
                        ),
                        (Field.METADATA_VALUE_RESTRICTION,),
                    ),
                ]
            )
            assert val_servicer_v1.Get.call_count == 1
            assert (
                val_servicer_v1.Get.call_args[0][0].entries
                == val_v1.GetRequest(
                    entries=(
                        val_v1.EntryRequest(
                            path="Vehicle.Speed",
                            view=View.METADATA,
                            fields=(Field.METADATA_DATA_TYPE,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.ADAS.ABS.IsActive",
                            view=View.METADATA,
                            fields=(Field.METADATA_DATA_TYPE,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.Cabin.Door.Row1.Left.Shade.Position",
                            view=View.METADATA,
                            fields=(Field.METADATA_DATA_TYPE,),
                        ),
                    )
                ).entries
            )
            assert (
                val_servicer_v1.Set.call_args[0][0].updates
                == val_v1.SetRequest(
                    updates=(
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Speed",
                                value=types_v1.Datapoint(float=42.0),
                            ),
                            fields=(types_v1.FIELD_VALUE,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.ABS.IsActive",
                                actuator_target=types_v1.Datapoint(bool=False),
                            ),
                            fields=(types_v1.FIELD_ACTUATOR_TARGET,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.CruiseControl.Error",
                                metadata=types_v1.Metadata(
                                    data_type=types_v1.DATA_TYPE_BOOLEAN,
                                    entry_type=types_v1.ENTRY_TYPE_SENSOR,
                                    description="Indicates if cruise control system incurred and error condition.",
                                    comment="No comment",
                                    deprecation="Never to be deprecated",
                                ),
                            ),
                            fields=(types_v1.FIELD_METADATA,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.CruiseControl.Error",
                                metadata=types_v1.Metadata(
                                    data_type=types_v1.DATA_TYPE_BOOLEAN
                                ),
                            ),
                            fields=(types_v1.FIELD_METADATA_DATA_TYPE,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.CruiseControl.Error",
                                metadata=types_v1.Metadata(
                                    description="Indicates if cruise control system incurred and error condition."
                                ),
                            ),
                            fields=(types_v1.FIELD_METADATA_DESCRIPTION,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.CruiseControl.Error",
                                metadata=types_v1.Metadata(
                                    entry_type=types_v1.ENTRY_TYPE_SENSOR
                                ),
                            ),
                            fields=(types_v1.FIELD_METADATA_ENTRY_TYPE,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.CruiseControl.Error",
                                metadata=types_v1.Metadata(comment="No comment"),
                            ),
                            fields=(types_v1.FIELD_METADATA_COMMENT,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.CruiseControl.Error",
                                metadata=types_v1.Metadata(
                                    deprecation="Never to be deprecated"
                                ),
                            ),
                            fields=(types_v1.FIELD_METADATA_DEPRECATION,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Cabin.Door.Row1.Left.Shade.Position",
                                metadata=types_v1.Metadata(unit="percent"),
                            ),
                            fields=(types_v1.FIELD_METADATA_UNIT,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Cabin.Door.Row1.Left.Shade.Position",
                                metadata=types_v1.Metadata(
                                    value_restriction=types_v1.ValueRestriction(
                                        unsigned=types_v1.ValueRestrictionUint(
                                            min=0, max=100
                                        ),
                                    )
                                ),
                            ),
                            fields=(types_v1.FIELD_METADATA_VALUE_RESTRICTION,),
                        ),
                    )
                ).updates
            )

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_set_some_updates_v2(
        self, unused_tcp_port, val_servicer_v2, val_servicer_v1
    ):
        val_servicer_v1.Get.return_value = val_v1.GetResponse(
            entries=(
                types_v1.DataEntry(
                    path="Vehicle.Speed",
                    metadata=types_v1.Metadata(data_type=types_v1.DATA_TYPE_FLOAT),
                ),
                types_v1.DataEntry(
                    path="Vehicle.ADAS.ABS.IsActive",
                    metadata=types_v1.Metadata(data_type=types_v1.DATA_TYPE_BOOLEAN),
                ),
            )
        )
        val_servicer_v2.PublishValue.return_value = val_v2.PublishValueResponse()
        _updates = [
            EntryUpdate(
                DataEntry("Vehicle.Speed", value=Datapoint(value=42.0)),
                (Field.VALUE,),
            ),
            EntryUpdate(
                DataEntry(
                    "Vehicle.ADAS.ABS.IsActive",
                    value=Datapoint(value=False),
                ),
                (Field.VALUE,),
            ),
        ]

        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            await client.set(
                updates=_updates,
                try_v2=True,
            )
            assert val_servicer_v1.Get.call_count == 1
            assert (
                val_servicer_v1.Get.call_args[0][0].entries
                == val_v1.GetRequest(
                    entries=(
                        val_v1.EntryRequest(
                            path="Vehicle.Speed",
                            view=View.METADATA,
                            fields=(Field.METADATA_DATA_TYPE,),
                        ),
                        val_v1.EntryRequest(
                            path="Vehicle.ADAS.ABS.IsActive",
                            view=View.METADATA,
                            fields=(Field.METADATA_DATA_TYPE,),
                        ),
                    )
                ).entries
            )

            expected_requests = [
                val_v2.PublishValueRequest(
                    signal_id=types_v2.SignalID(path="Vehicle.Speed"),
                    data_point=types_v2.Datapoint(value=types_v2.Value(float=42.0)),
                ),
                val_v2.PublishValueRequest(
                    signal_id=types_v2.SignalID(path="Vehicle.ADAS.ABS.IsActive"),
                    data_point=types_v2.Datapoint(value=types_v2.Value(bool=False)),
                ),
            ]

            assert val_servicer_v2.PublishValue.call_count == len(_updates)

            actual_requests = [
                call[0][0] for call in val_servicer_v2.PublishValue.call_args_list
            ]

            for actual_request, expected_request in zip(
                actual_requests, expected_requests
            ):
                assert actual_request == expected_request

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_set_some_updates_v2_target(
        self, unused_tcp_port, val_servicer_v2, val_servicer_v1
    ):
        """
        Similar test to above, but trying to update target values using v2
        which is not allowed
        """
        val_servicer_v1.Get.return_value = val_v1.GetResponse(
            entries=(
                types_v1.DataEntry(
                    path="Vehicle.Speed",
                    metadata=types_v1.Metadata(data_type=types_v1.DATA_TYPE_FLOAT),
                ),
                types_v1.DataEntry(
                    path="Vehicle.ADAS.ABS.IsActive",
                    metadata=types_v1.Metadata(data_type=types_v1.DATA_TYPE_BOOLEAN),
                ),
            )
        )
        _updates = [
            EntryUpdate(
                DataEntry("Vehicle.Speed", value=Datapoint(value=42.0)),
                (Field.ACTUATOR_TARGET,),
            ),
            EntryUpdate(
                DataEntry(
                    "Vehicle.ADAS.ABS.IsActive",
                    value=Datapoint(value=False),
                ),
                (Field.ACTUATOR_TARGET,),
            ),
        ]

        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            with pytest.raises(VSSClientError):
                await client.set(
                    updates=_updates,
                    try_v2=True,
                )
            assert val_servicer_v1.Get.call_count == 1
            assert val_servicer_v2.PublishValue.call_count == 0

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_set_no_updates_provided(
        self, unused_tcp_port, val_servicer_v1, val_servicer_v2
    ):
        val_servicer_v1.Set.side_effect = generate_error(
            grpc.StatusCode.INVALID_ARGUMENT, "No datapoints requested"
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            with pytest.raises(kuksa_client.grpc.VSSClientError) as exc_info:
                await client.set(updates=[])

            assert exc_info.value.args == kuksa_client.grpc.VSSClientError(error={
                'code': grpc.StatusCode.INVALID_ARGUMENT.value[0],
                'reason': grpc.StatusCode.INVALID_ARGUMENT.value[1],
                'message': 'No datapoints requested',
            }, errors=[]).args
            assert val_servicer_v1.Get.call_count == 0
            assert (
                val_servicer_v1.Set.call_args[0][0].updates
                == val_v1.SetRequest().updates
            )
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            with pytest.raises(kuksa_client.grpc.VSSClientError) as exc_info:
                await client.set(updates=[], try_v2=True)

            assert (
                exc_info.value.args
                == kuksa_client.grpc.VSSClientError(
                    error={
                        "code": grpc.StatusCode.INVALID_ARGUMENT.value[0],
                        "reason": grpc.StatusCode.INVALID_ARGUMENT.value[1],
                        "message": "No datapoints requested",
                    },
                    errors=[],
                ).args
            )
            assert val_servicer_v1.Get.call_count == 0
            assert val_servicer_v2.PublishValue.call_count == 0

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_set_nonexistent_entries_v1(self, unused_tcp_port, val_servicer_v1):
        error = types_v1.Error(
            code=404, reason="not_found", message="Does.Not.Exist not found"
        )
        errors = (types_v1.DataEntryError(path="Does.Not.Exist", error=error),)
        val_servicer_v1.Get.return_value = val_v1.GetResponse(
            error=error, errors=errors
        )
        val_servicer_v1.Set.return_value = val_v1.SetResponse(
            error=error, errors=errors
        )
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            with pytest.raises(VSSClientError):
                await client.set(
                    updates=(
                        EntryUpdate(
                            DataEntry("Does.Not.Exist", value=Datapoint(value=42.0)),
                            (Field.VALUE,),
                        ),
                    ),
                )

            assert val_servicer_v1.Get.call_count == 1
            assert val_servicer_v1.Set.call_count == 0
            with pytest.raises(VSSClientError):
                await client.set(
                    updates=(
                        EntryUpdate(
                            DataEntry(
                                "Does.Not.Exist",
                                value=Datapoint(value=42.0),
                                metadata=Metadata(data_type=DataType.FLOAT),
                            ),
                            (Field.VALUE,),
                        ),
                    ),
                )

            assert (
                val_servicer_v1.Get.call_count == 1
            )  # Get should'nt have been called again
            assert val_servicer_v1.Set.call_count == 1

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_set_nonexistent_entries_v2(
        self, unused_tcp_port, val_servicer_v2, val_servicer_v1
    ):
        error = types_v1.Error(
            code=404, reason="not_found", message="Does.Not.Exist not found"
        )
        errors = (types_v1.DataEntryError(path="Does.Not.Exist", error=error),)
        val_servicer_v1.Get.return_value = val_v1.GetResponse(
            error=error, errors=errors
        )
        val_servicer_v2.PublishValue.side_effect = generate_error(
            grpc.StatusCode.NOT_FOUND,
            "Does.Not.Exist not found",
        )
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            with pytest.raises(VSSClientError):
                await client.set(
                    updates=(
                        EntryUpdate(
                            DataEntry("Does.Not.Exist", value=Datapoint(value=42.0)),
                            (Field.VALUE,),
                        ),
                    ),
                    try_v2=True,
                )

            assert val_servicer_v1.Get.call_count == 1
            assert val_servicer_v2.PublishValue.call_count == 0
            with pytest.raises(VSSClientError):
                await client.set(
                    updates=(
                        EntryUpdate(
                            DataEntry(
                                "Does.Not.Exist",
                                value=Datapoint(value=42.0),
                                metadata=Metadata(data_type=DataType.FLOAT),
                            ),
                            (Field.VALUE,),
                        ),
                    ),
                    try_v2=True,
                )

            assert (
                val_servicer_v1.Get.call_count == 1
            )  # Get should'nt have been called again
            assert val_servicer_v2.PublishValue.call_count == 1

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_authorize_successful(self, unused_tcp_port, val_servicer_v1):
        val_servicer_v1.GetServerInfo.return_value = val_v1.GetServerInfoResponse(
            name="test_server", version="1.2.3"
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            # token from kuksa.val directory under jwt/provide-vehicle-speed.token
            token = ('eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJzdWIiOiJsb2NhbCBkZXYiLCJpc3MiOiJjcmVhdGVUb2'
                     'tlbi5weSIsImF1ZCI6WyJrdWtzYS52YWwiXSwiaWF0IjoxNTE2MjM5MDIyLCJleHAiOjE3NjcyMjU1OTksIn'
                     'Njb3BlIjoicmVhZDpWZWhpY2xlLldpZHRoIHByb3ZpZGU6VmVoaWNsZS5TcGVlZCJ9.w2c8xrYwBVgMav3f0S'
                     'e6E8H8E36Nd03rJiSS2A8s-CL3GtlwB7wVanjXHhppNsCdWym3tK4JwgslQdMQF-UL4hd7vzdtt-Mx6VjH_jO9'
                     'mDxz4Z0Uzw7aJtbtQSpi2h6kwceTVTllkbLRF7WRHWIpwzXFF9yZolX6lH-BE9xf1AB62d6icd9SKxFnVvYs3M'
                     'VK5D1xNmDNOmm-Fr0d2K604MmIIXGW5kPZJYIvBKO4NYRLklhJe47It_lGo3gnh1ppmzTOIo1kB4sDe55hplUCb'
                     'TCJVricpyQSgTYsf7aFRPK51XMRwwwJ8kShWeaTggMLKpv1W-9dhVWDk4isC8BxsOjaVloArausMmjLmTz6KwAsf'
                     'ARgfXtaCrMsESUBNXi5KIdAyHVXZpmERvc9yeYPcaWlknVFrFsHbV6bw4nwqBX-0Ubuga0NGNQDFKmyTKQrbuZmQ'
                     '3L9iipxY8_BOSCkdiYtWbE3lpplxpS_PaZl10KAaMmUfbcF9aYZunDEzEtoJgJe2EeGu3XDBtbyXVUKruImdSEdja'
                     'ImfUGQIWl5bMbVH4N4zK5jE45wT5FJiRUcA5pMN5wNmDYJJzgbxWNpYW40KZYPFc_7XUH8EZ2Cs69wDHam3ArkOs1'
                     'qMgMIoEPWVzHakjlVJfrPR9zQKxfirBtNNENIoHsBjJ_P4FEJCN4'
                     )
            bearer = ('Bearer eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJzdWIiOiJsb2NhbCBkZXYiLCJpc3MiOiJjcmVhdG'
                      'VUb2tlbi5weSIsImF1ZCI6WyJrdWtzYS52YWwiXSwiaWF0IjoxNTE2MjM5MDIyLCJleHAiOjE3NjcyMjU1OTks'
                      'InNjb3BlIjoicmVhZDpWZWhpY2xlLldpZHRoIHByb3ZpZGU6VmVoaWNsZS5TcGVlZCJ9.w2c8xrYwBVgMav3f0'
                      'Se6E8H8E36Nd03rJiSS2A8s-CL3GtlwB7wVanjXHhppNsCdWym3tK4JwgslQdMQF-UL4hd7vzdtt-Mx6VjH_jO'
                      '9mDxz4Z0Uzw7aJtbtQSpi2h6kwceTVTllkbLRF7WRHWIpwzXFF9yZolX6lH-BE9xf1AB62d6icd9SKxFnVvYs3M'
                      'VK5D1xNmDNOmm-Fr0d2K604MmIIXGW5kPZJYIvBKO4NYRLklhJe47It_lGo3gnh1ppmzTOIo1kB4sDe55hplUCbT'
                      'CJVricpyQSgTYsf7aFRPK51XMRwwwJ8kShWeaTggMLKpv1W-9dhVWDk4isC8BxsOjaVloArausMmjLmTz6KwAsf'
                      'ARgfXtaCrMsESUBNXi5KIdAyHVXZpmERvc9yeYPcaWlknVFrFsHbV6bw4nwqBX-0Ubuga0NGNQDFKmyTKQrbuZmQ'
                      '3L9iipxY8_BOSCkdiYtWbE3lpplxpS_PaZl10KAaMmUfbcF9aYZunDEzEtoJgJe2EeGu3XDBtbyXVUKruImdSEdj'
                      'aImfUGQIWl5bMbVH4N4zK5jE45wT5FJiRUcA5pMN5wNmDYJJzgbxWNpYW40KZYPFc_7XUH8EZ2Cs69wDHam3ArkO'
                      's1qMgMIoEPWVzHakjlVJfrPR9zQKxfirBtNNENIoHsBjJ_P4FEJCN4'
                      )
            success = await client.authorize(token)
            assert client.authorization_header == bearer
            assert success == "Authenticated"

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_authorize_unsuccessful(self, unused_tcp_port, val_servicer_v1):
        val_servicer_v1.GetServerInfo.side_effect = generate_error(
            grpc.StatusCode.UNAUTHENTICATED,
            'Invalid auth token: DecodeError("InvalidToken")',
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            with pytest.raises(VSSClientError):
                await client.authorize(token='')
            assert client.authorization_header is None

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_subscribe_some_entries_v1(
        self, mocker, unused_tcp_port, val_servicer_v1
    ):
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            responses = (
                # 1st response is subscription ack
                val_v1.SubscribeResponse(
                    updates=[
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Speed",
                                value=types_v1.Datapoint(
                                    timestamp=timestamp_pb2.Timestamp(
                                        seconds=1667837915, nanos=247307674
                                    ),
                                    float=42.0,
                                ),
                            ),
                            fields=(Field.VALUE,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.ABS.IsActive",
                                actuator_target=types_v1.Datapoint(bool=True),
                            ),
                            fields=(Field.ACTUATOR_TARGET,),
                        ),
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Chassis.Height",
                                metadata=types_v1.Metadata(
                                    data_type=types_v1.DATA_TYPE_UINT16,
                                ),
                            ),
                            fields=(Field.METADATA_DATA_TYPE,),
                        ),
                    ]
                ),
                # Remaining responses are actual events.
                val_v1.SubscribeResponse(
                    updates=[
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Speed",
                                value=types_v1.Datapoint(
                                    timestamp=timestamp_pb2.Timestamp(
                                        seconds=1667837912, nanos=247307674
                                    ),
                                    float=43.0,
                                ),
                            ),
                            fields=(Field.VALUE,),
                        ),
                    ]
                ),
                val_v1.SubscribeResponse(
                    updates=[
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.ADAS.ABS.IsActive",
                                actuator_target=types_v1.Datapoint(bool=False),
                            ),
                            fields=(Field.ACTUATOR_TARGET,),
                        ),
                    ]
                ),
                val_v1.SubscribeResponse(
                    updates=[
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Chassis.Height",
                                metadata=types_v1.Metadata(
                                    data_type=types_v1.DATA_TYPE_UINT8,
                                ),
                            ),
                            fields=(Field.METADATA_DATA_TYPE,),
                        ),
                    ]
                ),
            )
            val_servicer_v1.Subscribe.return_value = (
                response for response in responses
            )

            actual_responses = []
            async for updates in client.subscribe(
                entries=(
                    entry
                    for entry in (  # generator is intentional (Iterable)
                        EntryRequest(
                            # View is ignored, so we can provide any value
                            "Vehicle.Speed", View.FIELDS, (Field.VALUE,)
                        ),
                        EntryRequest(
                            "Vehicle.ADAS.ABS.IsActive",
                            View.UNSPECIFIED,
                            (Field.ACTUATOR_TARGET,),
                        ),
                        EntryRequest(
                            "Vehicle.Chassis.Height",
                            View.UNSPECIFIED,
                            (Field.METADATA_DATA_TYPE,),
                        ),
                    )
                )
            ):
                actual_responses.append(updates)

            assert actual_responses == [
                [
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.Speed",
                            value=Datapoint(
                                value=42.0,
                                timestamp=datetime.datetime(
                                    2022,
                                    11,
                                    7,
                                    16,
                                    18,
                                    35,
                                    247307,
                                    tzinfo=datetime.timezone.utc,
                                ),
                            ),
                        ),
                        fields=[Field.VALUE],
                    ),
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.ADAS.ABS.IsActive",
                            actuator_target=Datapoint(value=True),
                        ),
                        fields=[Field.ACTUATOR_TARGET],
                    ),
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.Chassis.Height",
                            metadata=Metadata(
                                data_type=DataType.UINT16,
                            ),
                        ),
                        fields=[Field.METADATA_DATA_TYPE],
                    ),
                ],
                [
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.Speed",
                            value=Datapoint(
                                value=43.0,
                                timestamp=datetime.datetime(
                                    2022,
                                    11,
                                    7,
                                    16,
                                    18,
                                    32,
                                    247307,
                                    tzinfo=datetime.timezone.utc,
                                ),
                            ),
                        ),
                        fields=[Field.VALUE],
                    )
                ],
                [
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.ADAS.ABS.IsActive",
                            actuator_target=Datapoint(
                                value=False,
                            ),
                        ),
                        fields=[Field.ACTUATOR_TARGET],
                    )
                ],
                [
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.Chassis.Height",
                            metadata=Metadata(
                                data_type=DataType.UINT8,
                            ),
                        ),
                        fields=[Field.METADATA_DATA_TYPE],
                    )
                ],
            ]

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_subscribe_some_entries_v2(
        self, mocker, unused_tcp_port, val_servicer_v2
    ):
        _entries: Dict[str, types_v2.Datapoint] = {
            "Vehicle.Speed": types_v2.Datapoint(
                timestamp=timestamp_pb2.Timestamp(seconds=1667837915, nanos=247307674),
                value=types_v2.Value(float=42.0),
            ),
            "Vehicle.ADAS.ABS.IsActive": types_v2.Datapoint(
                value=types_v2.Value(bool=True)
            ),
        }
        _entries_2: Dict[str, types_v2.Datapoint] = {
            "Vehicle.Speed": types_v2.Datapoint(
                timestamp=timestamp_pb2.Timestamp(seconds=1667837912, nanos=247307674),
                value=types_v2.Value(float=43.0),
            ),
            "Vehicle.ADAS.ABS.IsActive": types_v2.Datapoint(
                value=types_v2.Value(bool=False)
            ),
        }
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            responses = (
                val_v2.SubscribeResponse(entries=_entries),
                val_v2.SubscribeResponse(entries=_entries_2),
            )
            val_servicer_v2.Subscribe.return_value = (
                response for response in responses
            )

            actual_responses = []
            async for updates in client.v2_subscribe(
                paths=(
                    path
                    for path in (  # generator is intentional (Iterable)
                        "Vehicle.Speed",
                        "Vehicle.ADAS.ABS.IsActive",
                    )
                ),
            ):
                actual_responses.append(updates)

            assert actual_responses == [
                [
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.Speed",
                            value=Datapoint(
                                value=42.0,
                                timestamp=datetime.datetime(
                                    2022,
                                    11,
                                    7,
                                    16,
                                    18,
                                    35,
                                    247307,
                                    tzinfo=datetime.timezone.utc,
                                ),
                            ),
                        ),
                        fields=[Field.VALUE],
                    ),
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.ADAS.ABS.IsActive",
                            value=Datapoint(value=True),
                        ),
                        fields=[Field.VALUE],
                    ),
                ],
                [
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.Speed",
                            value=Datapoint(
                                value=43.0,
                                timestamp=datetime.datetime(
                                    2022,
                                    11,
                                    7,
                                    16,
                                    18,
                                    32,
                                    247307,
                                    tzinfo=datetime.timezone.utc,
                                ),
                            ),
                        ),
                        fields=[Field.VALUE],
                    ),
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.ADAS.ABS.IsActive",
                            value=Datapoint(value=False),
                        ),
                        fields=[Field.VALUE],
                    ),
                ],
            ]

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_subscribe_some_entries_v2_target(
        self, mocker, unused_tcp_port, val_servicer_v2
    ):
        """
        Similar to above but trying to subscribe to target values which is not possible using v2
        """
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            actual_responses = []

            with pytest.raises(VSSClientError):
                async for updates in client.v2_subscribe_actuation_requests(
                    paths=(
                        path
                        for path in (  # generator is intentional (Iterable)
                            "Vehicle.Speed",
                            "Vehicle.ADAS.ABS.IsActive",
                        )
                    ),
                ):
                    actual_responses.append(updates)

            assert not actual_responses

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_subscribe_no_entries_requested(
        self, mocker, unused_tcp_port, val_servicer_v1, val_servicer_v2
    ):
        val_servicer_v1.Subscribe.side_effect = generate_error(
            grpc.StatusCode.INVALID_ARGUMENT,
            "Subscription request must contain at least one entry.",
        )
        val_servicer_v2.Subscribe.side_effect = generate_error(
            grpc.StatusCode.INVALID_ARGUMENT,
            "Subscription request must contain at least one entry.",
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            with pytest.raises(VSSClientError):
                async for _ in client.subscribe(entries=()):
                    pass

        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            with pytest.raises(VSSClientError):
                async for _ in client.v2_subscribe(paths=()):
                    pass

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_subscribe_nonexistent_entries(
        self, mocker, unused_tcp_port, val_servicer_v1, val_servicer_v2
    ):
        val_servicer_v1.Subscribe.side_effect = generate_error(
            grpc.StatusCode.INVALID_ARGUMENT, "NotFound"
        )
        val_servicer_v2.Subscribe.side_effect = generate_error(
            grpc.StatusCode.INVALID_ARGUMENT, "NotFound"
        )

        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            with pytest.raises(VSSClientError):
                async for _ in client.subscribe(entries=(entry for entry in (  # generator is intentional (Iterable)
                    EntryRequest('Does.Not.Exist',
                                 View.CURRENT_VALUE, (Field.VALUE,)),
                ))):
                    pass

        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            with pytest.raises(VSSClientError):
                async for _ in client.v2_subscribe(paths=["Does.Not.Exist"]):
                    pass

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_get_server_info(self, unused_tcp_port, val_servicer_v1):
        val_servicer_v1.GetServerInfo.return_value = val_v1.GetServerInfoResponse(
            name="test_server", version="1.2.3"
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            server_info = await client.get_server_info()
            assert server_info == ServerInfo(
                name='test_server', version='1.2.3')

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_get_server_info_unavailable(self, unused_tcp_port, val_servicer_v1):
        val_servicer_v1.GetServerInfo.side_effect = generate_error(
            grpc.StatusCode.UNAVAILABLE, "Unavailable"
        )
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            with pytest.raises(VSSClientError):
                await client.get_server_info()


@pytest.mark.asyncio
class TestSubscriberManager:

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_add_subscriber_v1(self, mocker, unused_tcp_port, val_servicer_v1):
        async with VSSClient('127.0.0.1', unused_tcp_port, ensure_startup_connection=False) as client:
            subscriber_manager = SubscriberManager(client)
            responses = (
                # 1st response is subscription ack
                val_v1.SubscribeResponse(
                    updates=[
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Speed",
                                value=types_v1.Datapoint(
                                    timestamp=timestamp_pb2.Timestamp(
                                        seconds=1667837915, nanos=247307674
                                    ),
                                    float=42.0,
                                ),
                            ),
                            fields=(Field.VALUE,),
                        ),
                    ]
                ),
                # Remaining responses are actual events that should invoke callback.
                val_v1.SubscribeResponse(
                    updates=[
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Speed",
                                value=types_v1.Datapoint(
                                    timestamp=timestamp_pb2.Timestamp(
                                        seconds=1667837912, nanos=247307674
                                    ),
                                    float=43.0,
                                ),
                            ),
                            fields=(Field.VALUE,),
                        ),
                    ]
                ),
            )
            callback = mocker.Mock()
            val_servicer_v1.Subscribe.return_value = (
                response for response in responses
            )

            subscribe_response_stream = client.subscribe(
                entries=(
                    EntryRequest("Vehicle.Speed", View.CURRENT_VALUE, (Field.VALUE,)),
                )
            )
            sub_uid = await subscriber_manager.add_subscriber(
                subscribe_response_stream, callback=callback
            )

            assert isinstance(sub_uid, uuid.UUID)
            while callback.call_count < 1:
                await asyncio.sleep(0.01)
            actual_updates = [
                list(call_args[0][0]) for call_args in callback.call_args_list
            ]

            assert actual_updates == [
                [
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.Speed",
                            value=Datapoint(
                                value=43.0,
                                timestamp=datetime.datetime(
                                    2022,
                                    11,
                                    7,
                                    16,
                                    18,
                                    32,
                                    247307,
                                    tzinfo=datetime.timezone.utc,
                                ),
                            ),
                        ),
                        fields=[Field.VALUE],
                    )
                ],
            ]

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_remove_subscriber_v1(self, mocker, unused_tcp_port, val_servicer_v1):
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            subscriber_manager = SubscriberManager(client)
            responses = (
                val_v1.SubscribeResponse(
                    updates=[
                        val_v1.EntryUpdate(
                            entry=types_v1.DataEntry(
                                path="Vehicle.Speed",
                                value=types_v1.Datapoint(
                                    timestamp=timestamp_pb2.Timestamp(
                                        seconds=1667837915, nanos=247307674
                                    ),
                                    float=42.0,
                                ),
                            ),
                            fields=(Field.VALUE,),
                        ),
                    ]
                ),
            )
            val_servicer_v1.Subscribe.return_value = (
                response for response in responses
            )
            subscribe_response_stream = client.subscribe(
                entries=(
                    EntryRequest("Vehicle.Speed", View.CURRENT_VALUE, (Field.VALUE,)),
                )
            )
            sub_uid = await subscriber_manager.add_subscriber(
                subscribe_response_stream, callback=mocker.Mock()
            )
            subscriber = subscriber_manager.subscribers.get(sub_uid)

            await subscriber_manager.remove_subscriber(sub_uid)

            assert subscriber_manager.subscribers.get(sub_uid) is None
            assert subscriber.done()

            with pytest.raises(ValueError) as exc_info:
                await subscriber_manager.remove_subscriber(sub_uid)
            assert (
                exc_info.value.args[0] == f"Could not find subscription {str(sub_uid)}"
            )

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_add_subscriber_v2(self, mocker, unused_tcp_port, val_servicer_v2):
        _entries: Dict[str, types_v2.Datapoint] = {
            "Vehicle.Speed": types_v2.Datapoint(
                timestamp=timestamp_pb2.Timestamp(seconds=1667837915, nanos=247307674),
                value=types_v2.Value(float=42.0),
            ),
        }
        _entries_2: Dict[str, types_v2.Datapoint] = {
            "Vehicle.Speed": types_v2.Datapoint(
                timestamp=timestamp_pb2.Timestamp(seconds=1667837912, nanos=247307674),
                value=types_v2.Value(float=43.0),
            ),
        }
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            subscriber_manager = SubscriberManager(client)
            responses = (
                # 1st response is subscription ack
                val_v2.SubscribeResponse(entries=_entries),
                # Remaining responses are actual events that should invoke callback.
                val_v2.SubscribeResponse(entries=_entries_2),
            )
            callback = mocker.Mock()
            val_servicer_v2.Subscribe.return_value = (
                response for response in responses
            )

            subscribe_response_stream = client.v2_subscribe(
                paths=("Vehicle.Speed"),
            )
            sub_uid = await subscriber_manager.add_subscriber(
                subscribe_response_stream, callback=callback
            )

            assert isinstance(sub_uid, uuid.UUID)
            while callback.call_count < 1:
                await asyncio.sleep(0.01)
            actual_updates = [
                list(call_args[0][0]) for call_args in callback.call_args_list
            ]

            assert actual_updates == [
                [
                    EntryUpdate(
                        entry=DataEntry(
                            path="Vehicle.Speed",
                            value=Datapoint(
                                value=43.0,
                                timestamp=datetime.datetime(
                                    2022,
                                    11,
                                    7,
                                    16,
                                    18,
                                    32,
                                    247307,
                                    tzinfo=datetime.timezone.utc,
                                ),
                            ),
                        ),
                        fields=[Field.VALUE],
                    )
                ],
            ]

    @pytest.mark.usefixtures("mocked_databroker")
    async def test_remove_subscriber_v2(self, mocker, unused_tcp_port, val_servicer_v2):
        async with VSSClient(
            "127.0.0.1", unused_tcp_port, ensure_startup_connection=False
        ) as client:
            subscriber_manager = SubscriberManager(client)
            _entries: Dict[str, types_v2.Datapoint] = {
                "Vehicle.Speed": types_v2.Datapoint(
                    timestamp=timestamp_pb2.Timestamp(
                        seconds=1667837915, nanos=247307674
                    ),
                    value=types_v2.Value(float=42.0),
                ),
            }
            responses = (val_v2.SubscribeResponse(entries=_entries),)
            val_servicer_v2.Subscribe.return_value = (
                response for response in responses
            )
            subscribe_response_stream = client.v2_subscribe(
                paths=("Vehicle.Speed"),
            )
            sub_uid = await subscriber_manager.add_subscriber(
                subscribe_response_stream, callback=mocker.Mock()
            )
            subscriber = subscriber_manager.subscribers.get(sub_uid)

            await subscriber_manager.remove_subscriber(sub_uid)

            assert subscriber_manager.subscribers.get(sub_uid) is None
            assert subscriber.done()

            with pytest.raises(ValueError) as exc_info:
                await subscriber_manager.remove_subscriber(sub_uid)
            assert (
                exc_info.value.args[0] == f"Could not find subscription {str(sub_uid)}"
            )
